# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    overload,
)

from ...abc import Snowflake
from ...components import RoleSelectMenu
from ...enums import ComponentType, SelectDefaultValueType
from ...object import Object
from ...role import Role
from ...utils import MISSING
from .base import BaseSelect, P, SelectDefaultValueInputType, V_co, _create_decorator

if TYPE_CHECKING:
    from typing_extensions import Self

    from ..item import DecoratedItem, ItemCallbackType


__all__ = (
    "RoleSelect",
    "role_select",
)


class RoleSelect(BaseSelect[RoleSelectMenu, "Role", V_co]):
    """Represents a UI role select menu.

    This is usually represented as a drop down menu.

    In order to get the selected items that the user has chosen, use :attr:`.values`.

    .. versionadded:: 2.7

    Parameters
    ----------
    custom_id: :class:`str`
        The ID of the select menu that gets received during an interaction.
        If not given then one is generated for you.
    placeholder: Optional[:class:`str`]
        The placeholder text that is shown if nothing is selected, if any.
    min_values: :class:`int`
        The minimum number of items that must be chosen for this select menu.
        Defaults to 1 and must be between 1 and 25.
    max_values: :class:`int`
        The maximum number of items that must be chosen for this select menu.
        Defaults to 1 and must be between 1 and 25.
    disabled: :class:`bool`
        Whether the select is disabled.
    default_values: Optional[Sequence[Union[:class:`.Role`, :class:`.SelectDefaultValue`, :class:`.Object`]]]
        The list of values (roles) that are selected by default.
        If set, the number of items must be within the bounds set by ``min_values`` and ``max_values``.

        .. versionadded:: 2.10
    row: Optional[:class:`int`]
        The relative row this select menu belongs to. A Discord component can only have 5
        rows. By default, items are arranged automatically into those 5 rows. If you'd
        like to control the relative positioning of the row then passing an index is advised.
        For example, row=1 will show up before row=2. Defaults to ``None``, which is automatic
        ordering. The row number must be between 0 and 4 (i.e. zero indexed).

    Attributes
    ----------
    values: List[:class:`.Role`]
        A list of roles that have been selected by the user.
    """

    _default_value_type_map: ClassVar[
        Mapping[SelectDefaultValueType, Tuple[Type[Snowflake], ...]]
    ] = {
        SelectDefaultValueType.role: (Role, Object),
    }

    @overload
    def __init__(
        self: RoleSelect[None],
        *,
        custom_id: str = ...,
        placeholder: Optional[str] = None,
        min_values: int = 1,
        max_values: int = 1,
        disabled: bool = False,
        default_values: Optional[Sequence[SelectDefaultValueInputType[Role]]] = None,
        row: Optional[int] = None,
    ) -> None:
        ...

    @overload
    def __init__(
        self: RoleSelect[V_co],
        *,
        custom_id: str = ...,
        placeholder: Optional[str] = None,
        min_values: int = 1,
        max_values: int = 1,
        disabled: bool = False,
        default_values: Optional[Sequence[SelectDefaultValueInputType[Role]]] = None,
        row: Optional[int] = None,
    ) -> None:
        ...

    def __init__(
        self,
        *,
        custom_id: str = MISSING,
        placeholder: Optional[str] = None,
        min_values: int = 1,
        max_values: int = 1,
        disabled: bool = False,
        default_values: Optional[Sequence[SelectDefaultValueInputType[Role]]] = None,
        row: Optional[int] = None,
    ) -> None:
        super().__init__(
            RoleSelectMenu,
            ComponentType.role_select,
            custom_id=custom_id,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            disabled=disabled,
            default_values=default_values,
            row=row,
        )

    @classmethod
    def from_component(cls, component: RoleSelectMenu) -> Self:
        return cls(
            custom_id=component.custom_id,
            placeholder=component.placeholder,
            min_values=component.min_values,
            max_values=component.max_values,
            disabled=component.disabled,
            default_values=component.default_values,
            row=None,
        )


S_co = TypeVar("S_co", bound="RoleSelect", covariant=True)


@overload
def role_select(
    *,
    placeholder: Optional[str] = None,
    custom_id: str = ...,
    min_values: int = 1,
    max_values: int = 1,
    disabled: bool = False,
    default_values: Optional[Sequence[SelectDefaultValueInputType[Role]]] = None,
    row: Optional[int] = None,
) -> Callable[[ItemCallbackType[V_co, RoleSelect[V_co]]], DecoratedItem[RoleSelect[V_co]]]:
    ...


@overload
def role_select(
    cls: Callable[P, S_co], *_: P.args, **kwargs: P.kwargs
) -> Callable[[ItemCallbackType[V_co, S_co]], DecoratedItem[S_co]]:
    ...


def role_select(
    cls: Callable[..., S_co] = RoleSelect[Any], **kwargs: Any
) -> Callable[[ItemCallbackType[V_co, S_co]], DecoratedItem[S_co]]:
    """A decorator that attaches a role select menu to a component.

    The function being decorated should have three parameters, ``self`` representing
    the :class:`disnake.ui.View`, the :class:`disnake.ui.RoleSelect` that was
    interacted with, and the :class:`disnake.MessageInteraction`.

    In order to get the selected items that the user has chosen within the callback
    use :attr:`RoleSelect.values`.

    .. versionadded:: 2.7

    Parameters
    ----------
    cls: Callable[..., :class:`RoleSelect`]
        A callable (may be a :class:`RoleSelect` subclass) to create a new instance of this component.
        If provided, the other parameters described below do not apply.
        Instead, this decorator will accept the same keywords as the passed callable/class does.
    placeholder: Optional[:class:`str`]
        The placeholder text that is shown if nothing is selected, if any.
    custom_id: :class:`str`
        The ID of the select menu that gets received during an interaction.
        It is recommended not to set this parameter to prevent conflicts.
    row: Optional[:class:`int`]
        The relative row this select menu belongs to. A Discord component can only have 5
        rows. By default, items are arranged automatically into those 5 rows. If you'd
        like to control the relative positioning of the row then passing an index is advised.
        For example, row=1 will show up before row=2. Defaults to ``None``, which is automatic
        ordering. The row number must be between 0 and 4 (i.e. zero indexed).
    min_values: :class:`int`
        The minimum number of items that must be chosen for this select menu.
        Defaults to 1 and must be between 1 and 25.
    max_values: :class:`int`
        The maximum number of items that must be chosen for this select menu.
        Defaults to 1 and must be between 1 and 25.
    disabled: :class:`bool`
        Whether the select is disabled. Defaults to ``False``.
    default_values: Optional[Sequence[Union[:class:`.Role`, :class:`.SelectDefaultValue`, :class:`.Object`]]]
        The list of values (roles) that are selected by default.
        If set, the number of items must be within the bounds set by ``min_values`` and ``max_values``.

        .. versionadded:: 2.10
    """
    return _create_decorator(cls, **kwargs)



# SPDX-License-Identifier: MIT

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Optional

from .enums import SKUType, try_enum
from .flags import SKUFlags
from .iterators import SubscriptionIterator
from .mixins import Hashable
from .subscription import Subscription
from .utils import snowflake_time

if TYPE_CHECKING:
    from .abc import Snowflake, SnowflakeTime
    from .state import ConnectionState
    from .types.sku import SKU as SKUPayload


__all__ = ("SKU",)


class SKU(Hashable):
    """Represents an SKU.

    This can be retrieved using :meth:`Client.skus`.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two :class:`SKU`\\s are equal.

        .. describe:: x != y

            Checks if two :class:`SKU`\\s are not equal.

        .. describe:: hash(x)

            Returns the SKU's hash.

        .. describe:: str(x)

            Returns the SKU's name.

    .. versionadded:: 2.10

    Attributes
    ----------
    id: :class:`int`
        The SKU's ID.
    type: :class:`SKUType`
        The SKU's type.
    application_id: :class:`int`
        The parent application's ID.
    name: :class:`str`
        The SKU's name.
    slug: :class:`str`
        The SKU's URL slug, system-generated based on :attr:`name`.
    """

    __slots__ = ("_state", "id", "type", "application_id", "name", "slug", "_flags")

    def __init__(self, *, data: SKUPayload, state: ConnectionState) -> None:
        self._state: ConnectionState = state
        self.id: int = int(data["id"])
        self.type: SKUType = try_enum(SKUType, data["type"])
        self.application_id: int = int(data["application_id"])
        self.name: str = data["name"]
        self.slug: str = data["slug"]
        self._flags: int = data.get("flags", 0)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<SKU id={self.id!r} type={self.type!r} name={self.name!r}>"

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns the SKU's creation time in UTC."""
        return snowflake_time(self.id)

    @property
    def flags(self) -> SKUFlags:
        """:class:`SKUFlags`: Returns the SKU's flags."""
        return SKUFlags._from_value(self._flags)

    async def subscriptions(
        self,
        user: Snowflake,
        *,
        limit: Optional[int] = 50,
        before: Optional[SnowflakeTime] = None,
        after: Optional[SnowflakeTime] = None,
    ) -> SubscriptionIterator:
        """|coro|

        Retrieves an :class:`.AsyncIterator` that enables receiving subscriptions for the SKU.

        All parameters, except ``user``, are optional.

        Parameters
        ----------
        user: :class:`abc.Snowflake`
            The user to retrieve subscriptions for.
        limit: Optional[:class:`int`]
            The number of subscriptions to retrieve.
            If ``None``, retrieves every subscription.
            Note, however, that this would make it a slow operation.
            Defaults to ``50``.
        before: Union[:class:`.abc.Snowflake`, :class:`datetime.datetime`]
            Retrieves subscriptions created before this date or object.
            If a datetime is provided, it is recommended to use a UTC aware datetime.
            If the datetime is naive, it is assumed to be local time.
        after: Union[:class:`.abc.Snowflake`, :class:`datetime.datetime`]
            Retrieve subscriptions created after this date or object.
            If a datetime is provided, it is recommended to use a UTC aware datetime.
            If the datetime is naive, it is assumed to be local time.

        Raises
        ------
        HTTPException
            Retrieving the subscriptions failed.

        Yields
        ------
        :class:`.Subscription`
            The subscriptions for the given parameters.
        """
        return SubscriptionIterator(
            self.id,
            state=self._state,
            user_id=user.id,
            limit=limit,
            before=before,
            after=after,
        )

    async def fetch_subscription(self, subscription_id: int, /) -> Subscription:
        """|coro|

        Retrieve a subscription for this SKU given its ID.

        Raises
        ------
        NotFound
            The subscription does not exist.
        HTTPException
            Retrieving the subscription failed.
        """
        data = await self._state.http.get_subscription(self.id, subscription_id)
        return Subscription(data=data, state=self._state)



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional

from ..enums import ComponentType
from ..message import Message
from ..utils import cached_slot_property
from .base import ClientT, Interaction

if TYPE_CHECKING:
    from ..state import ConnectionState
    from ..types.interactions import (
        ModalInteraction as ModalInteractionPayload,
        ModalInteractionActionRow as ModalInteractionActionRowPayload,
        ModalInteractionComponentData as ModalInteractionComponentDataPayload,
        ModalInteractionData as ModalInteractionDataPayload,
    )

__all__ = ("ModalInteraction", "ModalInteractionData")


class ModalInteraction(Interaction[ClientT]):
    """Represents an interaction with a modal.

    .. versionadded:: 2.4

    Attributes
    ----------
    id: :class:`int`
        The interaction's ID.
    type: :class:`InteractionType`
        The interaction type.
    application_id: :class:`int`
        The application ID that the interaction was for.
    token: :class:`str`
        The token to continue the interaction.
        These are valid for 15 minutes.
    guild_id: Optional[:class:`int`]
        The guild ID the interaction was sent from.
    channel: Union[:class:`abc.GuildChannel`, :class:`Thread`, :class:`abc.PrivateChannel`, :class:`PartialMessageable`]
        The channel the interaction was sent from.

        Note that due to a Discord limitation, DM channels
        may not contain recipient information.
        Unknown channel types will be :class:`PartialMessageable`.

        .. versionchanged:: 2.10
            If the interaction was sent from a thread and the bot cannot normally access the thread,
            this is now a proper :class:`Thread` object.
            Private channels are now proper :class:`DMChannel`/:class:`GroupChannel`
            objects instead of :class:`PartialMessageable`.

        .. note::
            If you want to compute the interaction author's or bot's permissions in the channel,
            consider using :attr:`permissions` or :attr:`app_permissions`.

    author: Union[:class:`User`, :class:`Member`]
        The user or member that sent the interaction.

        .. note::
            In scenarios where an interaction occurs in a guild but :attr:`.guild` is unavailable,
            such as with user-installed applications in guilds, some attributes of :class:`Member`\\s
            that depend on the guild/role cache will not work due to an API limitation.
            This includes :attr:`~Member.roles`, :attr:`~Member.top_role`, :attr:`~Member.role_icon`,
            and :attr:`~Member.guild_permissions`.
    locale: :class:`Locale`
        The selected language of the interaction's author.

        .. versionchanged:: 2.5
            Changed to :class:`Locale` instead of :class:`str`.

    guild_locale: Optional[:class:`Locale`]
        The selected language of the interaction's guild.
        This value is only meaningful in guilds with ``COMMUNITY`` feature and receives a default value otherwise.
        If the interaction was in a DM, then this value is ``None``.

        .. versionchanged:: 2.5
            Changed to :class:`Locale` instead of :class:`str`.

    client: :class:`Client`
        The interaction client.
    entitlements: List[:class:`Entitlement`]
        The entitlements for the invoking user and guild,
        representing access to an application subscription.

        .. versionadded:: 2.10

    authorizing_integration_owners: :class:`AuthorizingIntegrationOwners`
        Details about the authorizing user/guild for the application installation
        related to the interaction.

        .. versionadded:: 2.10

    context: :class:`InteractionContextTypes`
        The context where the interaction was triggered from.

        This is a flag object, with exactly one of the flags set to ``True``.
        To check whether an interaction originated from e.g. a :attr:`~InteractionContextTypes.guild`
        context, you can use ``if interaction.context.guild:``.

        .. versionadded:: 2.10

    data: :class:`ModalInteractionData`
        The wrapped interaction data.
    message: Optional[:class:`Message`]
        The message that this interaction's modal originated from,
        if the modal was sent in response to a component interaction.

        .. versionadded:: 2.5
    """

    __slots__ = ("message", "_cs_text_values")

    def __init__(self, *, data: ModalInteractionPayload, state: ConnectionState) -> None:
        super().__init__(data=data, state=state)
        self.data: ModalInteractionData = ModalInteractionData(data=data["data"])

        if message_data := data.get("message"):
            message = Message(state=self._state, channel=self.channel, data=message_data)
        else:
            message = None
        self.message: Optional[Message] = message

    def walk_raw_components(self) -> Generator[ModalInteractionComponentDataPayload, None, None]:
        """Returns a generator that yields raw component data from action rows one by one, as provided by Discord.
        This does not contain all fields of the components due to API limitations.

        .. versionadded:: 2.6

        Returns
        -------
        Generator[:class:`dict`, None, None]
        """
        for action_row in self.data.components:
            yield from action_row["components"]

    @cached_slot_property("_cs_text_values")
    def text_values(self) -> Dict[str, str]:
        """Dict[:class:`str`, :class:`str`]: Returns the text values the user has entered in the modal.
        This is a dict of the form ``{custom_id: value}``.
        """
        text_input_type = ComponentType.text_input.value
        return {
            component["custom_id"]: component.get("value") or ""
            for component in self.walk_raw_components()
            if component.get("type") == text_input_type
        }

    @property
    def custom_id(self) -> str:
        """:class:`str`: The custom ID of the modal."""
        return self.data.custom_id


class ModalInteractionData(Dict[str, Any]):
    """Represents the data of an interaction with a modal.

    .. versionadded:: 2.4

    Attributes
    ----------
    custom_id: :class:`str`
        The custom ID of the modal.
    components: List[:class:`dict`]
        The raw component data of the modal interaction, as provided by Discord.
        This does not contain all fields of the components due to API limitations.

        .. versionadded:: 2.6
    """

    __slots__ = ("custom_id", "components")

    def __init__(self, *, data: ModalInteractionDataPayload) -> None:
        super().__init__(data)
        self.custom_id: str = data["custom_id"]
        # This uses a stripped-down action row TypedDict, as we only receive
        # partial data from the API, generally only containing `type`, `custom_id`,
        # and relevant fields like a select's `values`.
        self.components: List[ModalInteractionActionRowPayload] = data["components"]

    def __repr__(self) -> str:
        return f"<ModalInteractionData custom_id={self.custom_id!r} components={self.components!r}>"



# SPDX-License-Identifier: MIT

from datetime import datetime, timezone

import pytest

from disnake import Object

snowflake = 881536165478499999  # date/time of first commit


def test_init() -> None:
    with pytest.raises(
        TypeError, match=r"id parameter must be convertable to int not <class 'str'>"
    ):
        Object("hi")


def test_compare() -> None:
    assert Object(42) == Object(42)
    assert Object(42) != Object(43)


def test_hash() -> None:
    assert hash(Object(snowflake)) == 210174600000


def test_created_at() -> None:
    assert Object(snowflake).created_at == datetime(2021, 8, 29, 13, 50, 0, tzinfo=timezone.utc)



# SPDX-License-Identifier: MIT

from disnake import DiscordWarning

__all__ = ("MessageContentPrefixWarning",)


class MessageContentPrefixWarning(DiscordWarning):
    """Warning for invalid prefixes without message content."""

    pass



# SPDX-License-Identifier: MIT

from __future__ import annotations

import datetime
from typing import (
    TYPE_CHECKING,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Optional,
    Sequence,
    Type,
    Union,
    overload,
)

from .enums import (
    AutoModActionType,
    AutoModEventType,
    AutoModTriggerType,
    enum_if_int,
    try_enum,
    try_enum_to_int,
)
from .flags import AutoModKeywordPresets
from .utils import MISSING, _get_as_snowflake, snowflake_time

if TYPE_CHECKING:
    from typing_extensions import Self

    from .abc import Snowflake
    from .guild import Guild, GuildChannel
    from .member import Member
    from .message import Message
    from .role import Role
    from .threads import Thread
    from .types.automod import (
        AutoModAction as AutoModActionPayload,
        AutoModActionMetadata,
        AutoModBlockMessageActionMetadata,
        AutoModRule as AutoModRulePayload,
        AutoModSendAlertActionMetadata,
        AutoModTimeoutActionMetadata,
        AutoModTriggerMetadata as AutoModTriggerMetadataPayload,
        EditAutoModRule as EditAutoModRulePayload,
    )
    from .types.gateway import AutoModerationActionExecutionEvent

__all__ = (
    "AutoModAction",
    "AutoModBlockMessageAction",
    "AutoModSendAlertAction",
    "AutoModTimeoutAction",
    "AutoModTriggerMetadata",
    "AutoModRule",
    "AutoModActionExecution",
)


class AutoModAction:
    """A base class for auto moderation actions.

    This class is not meant to be instantiated by the user.
    The user-constructible subclasses are:

    - :class:`AutoModBlockMessageAction`
    - :class:`AutoModSendAlertAction`
    - :class:`AutoModTimeoutAction`

    Actions received from the API may be of this type
    (and not one of the subtypes above) if the action type is not implemented yet.

    .. versionadded:: 2.6

    Attributes
    ----------
    type: :class:`AutoModActionType`
        The action type.
    """

    __slots__ = ("type", "_metadata")

    def __init__(
        self,
        *,
        type: AutoModActionType,
    ) -> None:
        self.type: AutoModActionType = enum_if_int(AutoModActionType, type)
        self._metadata: AutoModActionMetadata = {}

    def __repr__(self) -> str:
        return f"<{type(self).__name__} type={self.type!r}>"

    @classmethod
    def _from_dict(cls, data: AutoModActionPayload) -> Self:
        self = cls.__new__(cls)

        self.type = try_enum(AutoModActionType, data["type"])
        self._metadata = data.get("metadata", {})

        return self

    def to_dict(self) -> AutoModActionPayload:
        return {
            "type": self.type.value,
            "metadata": self._metadata,
        }


class AutoModBlockMessageAction(AutoModAction):
    """Represents an auto moderation action that blocks content from being sent.

    .. versionadded:: 2.6

    Parameters
    ----------
    custom_message: Optional[:class:`str`]
        The custom message to show to the user when the rule is triggered.
        Maximum length is 150 characters.

        .. versionadded:: 2.9

    Attributes
    ----------
    type: :class:`AutoModActionType`
        The action type.
        Always set to :attr:`~AutoModActionType.block_message`.
    """

    __slots__ = ()

    _metadata: AutoModBlockMessageActionMetadata

    def __init__(self, custom_message: Optional[str] = None) -> None:
        super().__init__(type=AutoModActionType.block_message)

        if custom_message is not None:
            self._metadata["custom_message"] = custom_message

    @property
    def custom_message(self) -> Optional[str]:
        """Optional[:class:`str`]: The custom message to show to the user when the rule is triggered.

        .. versionadded:: 2.9
        """
        return self._metadata.get("custom_message")

    def __repr__(self) -> str:
        return f"<{type(self).__name__} custom_message={self.custom_message!r}>"


class AutoModSendAlertAction(AutoModAction):
    """Represents an auto moderation action that sends an alert to a channel.

    .. versionadded:: 2.6

    Parameters
    ----------
    channel: :class:`abc.Snowflake`
        The channel to send an alert in when the rule is triggered.

    Attributes
    ----------
    type: :class:`AutoModActionType`
        The action type.
        Always set to :attr:`~AutoModActionType.send_alert_message`.
    """

    __slots__ = ()

    _metadata: AutoModSendAlertActionMetadata

    def __init__(self, channel: Snowflake) -> None:
        super().__init__(type=AutoModActionType.send_alert_message)

        self._metadata["channel_id"] = channel.id

    @property
    def channel_id(self) -> int:
        """:class:`int`: The channel ID to send an alert in when the rule is triggered."""
        return int(self._metadata["channel_id"])

    def __repr__(self) -> str:
        return f"<{type(self).__name__} channel_id={self.channel_id!r}>"


class AutoModTimeoutAction(AutoModAction):
    """Represents an auto moderation action that times out the user.

    .. versionadded:: 2.6

    Parameters
    ----------
    duration: Union[:class:`int`, :class:`datetime.timedelta`]
        The duration (seconds or timedelta) for which to timeout the user when the rule is triggered.

    Attributes
    ----------
    type: :class:`AutoModActionType`
        The action type.
        Always set to :attr:`~AutoModActionType.timeout`.
    """

    __slots__ = ()

    _metadata: AutoModTimeoutActionMetadata

    def __init__(self, duration: Union[int, datetime.timedelta]) -> None:
        super().__init__(type=AutoModActionType.timeout)

        if isinstance(duration, datetime.timedelta):
            duration = int(duration.total_seconds())
        self._metadata["duration_seconds"] = duration

    @property
    def duration(self) -> int:
        """:class:`int`: The duration (in seconds) for which to timeout
        the user when the rule is triggered.
        """
        return self._metadata["duration_seconds"]

    def __repr__(self) -> str:
        return f"<{type(self).__name__} duration={self.duration!r}>"


class AutoModTriggerMetadata:
    """Metadata for an auto moderation trigger.

    Based on the trigger type, different fields can be used with various limits:

    .. csv-table::
        :header: "Trigger Type", ``keyword_filter``, ``regex_patterns``, ``presets``, ``allow_list``, ``mention_total_limit``, ``mention_raid_protection_enabled``

        :attr:`~AutoModTriggerType.keyword`,        ✅ (x1000), ✅ (x10), ❌, ✅ (x100),  ❌, ❌
        :attr:`~AutoModTriggerType.spam`,           ❌,         ❌,       ❌, ❌,         ❌, ❌
        :attr:`~AutoModTriggerType.keyword_preset`, ❌,         ❌,       ✅, ✅ (x1000), ❌, ❌
        :attr:`~AutoModTriggerType.mention_spam`,   ❌,         ❌,       ❌, ❌,         ✅, ✅

    .. versionadded:: 2.6

    Attributes
    ----------
    keyword_filter: Optional[Sequence[:class:`str`]]
        The list of keywords to check for, up to 1000 keywords. Used with :attr:`AutoModTriggerType.keyword`.

        See :ddocs:`api docs <resources/auto-moderation#auto-moderation-rule-object-keyword-matching-strategies>`
        for details about how keyword matching works.
        Each keyword must be 60 characters or less.

    regex_patterns: Optional[Sequence[:class:`str`]]
        The list of regular expressions to check for. Used with :attr:`AutoModTriggerType.keyword`.

        A maximum of 10 regexes can be added, each with up to 260 characters.

        .. note::

            Only Rust flavored regex is currently supported, which can be tested in online editors such as `Rustexp <https://rustexp.lpil.uk/>`__.

        .. versionadded:: 2.7

    presets: Optional[:class:`AutoModKeywordPresets`]
        The keyword presets. Used with :attr:`AutoModTriggerType.keyword_preset`.

    allow_list: Optional[Sequence[:class:`str`]]
        The keywords that should be exempt from a preset.
        Used with :attr:`AutoModTriggerType.keyword` (up to 100 exemptions) and :attr:`AutoModTriggerType.keyword_preset` (up to 1000 exemptions).

        Each keyword must be 60 characters or less.

    mention_total_limit: Optional[:class:`int`]
        The maximum number of mentions (members + roles) allowed, between 1 and 50. Used with :attr:`AutoModTriggerType.mention_spam`.

    mention_raid_protection_enabled: Optional[:class:`bool`]
        Whether to automatically detect mention raids. Used with :attr:`AutoModTriggerType.mention_spam`.

        Defaults to ``False``.

        .. versionadded:: 2.9
    """

    __slots__ = (
        "keyword_filter",
        "regex_patterns",
        "presets",
        "allow_list",
        "mention_total_limit",
        "mention_raid_protection_enabled",
    )

    @overload
    def __init__(
        self,
        *,
        keyword_filter: Optional[Sequence[str]],
        regex_patterns: Optional[Sequence[str]] = None,
        allow_list: Optional[Sequence[str]] = None,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        *,
        keyword_filter: Optional[Sequence[str]] = None,
        regex_patterns: Optional[Sequence[str]],
        allow_list: Optional[Sequence[str]] = None,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        *,
        presets: AutoModKeywordPresets,
        allow_list: Optional[Sequence[str]] = None,
    ) -> None:
        ...

    @overload
    def __init__(
        self, *, mention_total_limit: int, mention_raid_protection_enabled: bool = False
    ) -> None:
        ...

    def __init__(
        self,
        *,
        keyword_filter: Optional[Sequence[str]] = None,
        regex_patterns: Optional[Sequence[str]] = None,
        presets: Optional[AutoModKeywordPresets] = None,
        allow_list: Optional[Sequence[str]] = None,
        mention_total_limit: Optional[int] = None,
        mention_raid_protection_enabled: Optional[bool] = None,
    ) -> None:
        self.keyword_filter: Optional[Sequence[str]] = keyword_filter
        self.regex_patterns: Optional[Sequence[str]] = regex_patterns
        self.presets: Optional[AutoModKeywordPresets] = presets
        self.allow_list: Optional[Sequence[str]] = allow_list
        self.mention_total_limit: Optional[int] = mention_total_limit
        self.mention_raid_protection_enabled: Optional[bool] = mention_raid_protection_enabled

    def with_changes(
        self,
        *,
        keyword_filter: Optional[Sequence[str]] = MISSING,
        regex_patterns: Optional[Sequence[str]] = MISSING,
        presets: Optional[AutoModKeywordPresets] = MISSING,
        allow_list: Optional[Sequence[str]] = MISSING,
        mention_total_limit: Optional[int] = MISSING,
        mention_raid_protection_enabled: Optional[bool] = MISSING,
    ) -> Self:
        """Returns a new instance with the given changes applied.
        All other fields will be kept intact.

        Returns
        -------
        :class:`AutoModTriggerMetadata`
            The new metadata instance.
        """
        return self.__class__(  # type: ignore  # call doesn't match any overloads
            keyword_filter=self.keyword_filter if keyword_filter is MISSING else keyword_filter,
            regex_patterns=self.regex_patterns if regex_patterns is MISSING else regex_patterns,
            presets=self.presets if presets is MISSING else presets,
            allow_list=self.allow_list if allow_list is MISSING else allow_list,
            mention_total_limit=(
                self.mention_total_limit if mention_total_limit is MISSING else mention_total_limit
            ),
            mention_raid_protection_enabled=(
                self.mention_raid_protection_enabled
                if mention_raid_protection_enabled is MISSING
                else mention_raid_protection_enabled
            ),
        )

    @classmethod
    def _from_dict(cls, data: AutoModTriggerMetadataPayload) -> Self:
        if (presets_data := data.get("presets")) is not None:
            presets = AutoModKeywordPresets._from_values(presets_data)
        else:
            presets = None

        return cls(  # type: ignore  # call doesn't match any overloads
            keyword_filter=data.get("keyword_filter"),
            regex_patterns=data.get("regex_patterns"),
            presets=presets,
            allow_list=data.get("allow_list"),
            mention_total_limit=data.get("mention_total_limit"),
            mention_raid_protection_enabled=data.get("mention_raid_protection_enabled"),
        )

    def to_dict(self) -> AutoModTriggerMetadataPayload:
        data: AutoModTriggerMetadataPayload = {}
        if self.keyword_filter is not None:
            data["keyword_filter"] = list(self.keyword_filter)
        if self.regex_patterns is not None:
            data["regex_patterns"] = list(self.regex_patterns)
        if self.presets is not None:
            data["presets"] = self.presets.values  # type: ignore  # `values` contains ints instead of preset literal values
        if self.allow_list is not None:
            data["allow_list"] = list(self.allow_list)
        if self.mention_total_limit is not None:
            data["mention_total_limit"] = self.mention_total_limit
        if self.mention_raid_protection_enabled is not None:
            data["mention_raid_protection_enabled"] = self.mention_raid_protection_enabled
        return data

    def __repr__(self) -> str:
        s = f"<{type(self).__name__}"
        if self.keyword_filter is not None:
            s += f" keyword_filter={self.keyword_filter!r}"
        if self.regex_patterns is not None:
            s += f" regex_patterns={self.regex_patterns!r}"
        if self.presets is not None:
            s += f" presets={self.presets!r}"
        if self.allow_list is not None:
            s += f" allow_list={self.allow_list!r}"
        if self.mention_total_limit is not None:
            s += f" mention_total_limit={self.mention_total_limit!r}"
        if self.mention_raid_protection_enabled is not None:
            s += f" mention_raid_protection_enabled={self.mention_raid_protection_enabled!r}"
        return f"{s}>"


class AutoModRule:
    """Represents an auto moderation rule.

    .. versionadded:: 2.6

    Attributes
    ----------
    id: :class:`int`
        The rule ID.
    name: :class:`str`
        The rule name.
    enabled: :class:`bool`
        Whether this rule is enabled.
    guild: :class:`Guild`
        The guild of the rule.
    creator_id: :class:`int`
        The rule creator's ID. See also :attr:`.creator`.
    event_type: :class:`AutoModEventType`
        The event type this rule is applied to.
    trigger_type: :class:`AutoModTriggerType`
        The type of trigger that determines whether this rule's actions should run for a specific event.
    trigger_metadata: :class:`AutoModTriggerMetadata`
        Additional metadata associated with this rule's :attr:`.trigger_type`.
    exempt_role_ids: FrozenSet[:class:`int`]
        The role IDs that are exempt from this rule.
    exempt_channel_ids: FrozenSet[:class:`int`]
        The channel IDs that are exempt from this rule.
    """

    __slots__ = (
        "id",
        "name",
        "enabled",
        "guild",
        "creator_id",
        "event_type",
        "trigger_type",
        "trigger_metadata",
        "_actions",
        "exempt_role_ids",
        "exempt_channel_ids",
    )

    def __init__(self, *, data: AutoModRulePayload, guild: Guild) -> None:
        self.guild: Guild = guild

        self.id: int = int(data["id"])
        self.name: str = data["name"]
        self.enabled: bool = data["enabled"]
        self.creator_id: int = int(data["creator_id"])
        self.event_type: AutoModEventType = try_enum(AutoModEventType, data["event_type"])
        self.trigger_type: AutoModTriggerType = try_enum(AutoModTriggerType, data["trigger_type"])
        self._actions: List[AutoModAction] = [
            _automod_action_factory(action) for action in data["actions"]
        ]
        self.trigger_metadata: AutoModTriggerMetadata = AutoModTriggerMetadata._from_dict(
            data.get("trigger_metadata", {})
        )
        self.exempt_role_ids: FrozenSet[int] = (
            frozenset(map(int, exempt_roles))
            if (exempt_roles := data.get("exempt_roles"))
            else frozenset()
        )
        self.exempt_channel_ids: FrozenSet[int] = (
            frozenset(map(int, exempt_channels))
            if (exempt_channels := data.get("exempt_channels"))
            else frozenset()
        )

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns the rule's creation time in UTC.

        .. versionadded:: 2.10
        """
        return snowflake_time(self.id)

    @property
    def actions(self) -> List[AutoModAction]:
        """List[Union[:class:`AutoModBlockMessageAction`, :class:`AutoModSendAlertAction`, :class:`AutoModTimeoutAction`, :class:`AutoModAction`]]:
        The list of actions that will execute if a matching event triggered this rule.
        """
        return list(self._actions)  # return a copy

    @property
    def creator(self) -> Optional[Member]:
        """Optional[:class:`Member`]: The guild member that created this rule.
        May be ``None`` if the member cannot be found. See also :attr:`.creator_id`.
        """
        return self.guild.get_member(self.creator_id)

    @property
    def exempt_roles(self) -> List[Role]:
        """List[:class:`Role`]: The list of roles that are exempt from this rule."""
        return list(filter(None, map(self.guild.get_role, self.exempt_role_ids)))

    @property
    def exempt_channels(self) -> List[GuildChannel]:
        """List[:class:`abc.GuildChannel`]: The list of channels that are exempt from this rule."""
        return list(filter(None, map(self.guild.get_channel, self.exempt_channel_ids)))

    def __repr__(self) -> str:
        return (
            f"<AutoModRule id={self.id!r} name={self.name!r} enabled={self.enabled!r}"
            f" creator={self.creator!r} event_type={self.event_type!r} trigger_type={self.trigger_type!r}"
            f" actions={self._actions!r} exempt_roles={self.exempt_role_ids!r} exempt_channels={self.exempt_channel_ids!r}"
            f" trigger_metadata={self.trigger_metadata!r}>"
        )

    async def edit(
        self,
        *,
        name: str = MISSING,
        event_type: AutoModEventType = MISSING,
        trigger_metadata: AutoModTriggerMetadata = MISSING,
        actions: Sequence[AutoModAction] = MISSING,
        enabled: bool = MISSING,
        exempt_roles: Optional[Iterable[Snowflake]] = MISSING,
        exempt_channels: Optional[Iterable[Snowflake]] = MISSING,
        reason: Optional[str] = None,
    ) -> AutoModRule:
        """|coro|

        Edits the auto moderation rule.

        You must have :attr:`.Permissions.manage_guild` permission to do this.

        All fields are optional.

        .. versionchanged:: 2.9
            Now raises a :exc:`TypeError` if given ``actions`` have an invalid type.

        Examples
        --------
        Edit name and enable rule:

        .. code-block:: python3

            await rule.edit(name="cool new rule", enabled=True)

        Add an action:

        .. code-block:: python3

            await rule.edit(
                actions=rule.actions + [AutoModTimeoutAction(3600)],
            )

        Add a keyword to a keyword filter rule:

        .. code-block:: python3

            meta = rule.trigger_metadata
            await rule.edit(
                trigger_metadata=meta.with_changes(
                    keyword_filter=meta.keyword_filter + ["stuff"],
                ),
            )

        Parameters
        ----------
        name: :class:`str`
            The rule's new name.
        event_type: :class:`AutoModEventType`
            The rule's new event type.
        trigger_metadata: :class:`AutoModTriggerMetadata`
            The rule's new associated trigger metadata.
        actions: Sequence[Union[:class:`AutoModBlockMessageAction`, :class:`AutoModSendAlertAction`, :class:`AutoModTimeoutAction`, :class:`AutoModAction`]]
            The rule's new actions.
            If provided, must contain at least one action.
        enabled: :class:`bool`
            Whether to enable the rule.
        exempt_roles: Optional[Iterable[:class:`abc.Snowflake`]]
            The rule's new exempt roles, up to 20.
            If ``[]`` or ``None`` is passed then all role exemptions are removed.
        exempt_channels: Optional[Iterable[:class:`abc.Snowflake`]]
            The rule's new exempt channels, up to 50.
            Can also include categories, in which case all channels inside that category will be exempt.
            If ``[]`` or ``None`` is passed then all channel exemptions are removed.
        reason: Optional[:class:`str`]
            The reason for editing the rule. Shows up on the audit log.

        Raises
        ------
        ValueError
            When editing the list of actions, at least one action must be provided.
        TypeError
            The specified ``actions`` are of an invalid type.
        Forbidden
            You do not have proper permissions to edit the rule.
        NotFound
            The rule does not exist.
        HTTPException
            Editing the rule failed.

        Returns
        -------
        :class:`AutoModRule`
            The newly updated auto moderation rule.
        """
        payload: EditAutoModRulePayload = {}

        if name is not MISSING:
            payload["name"] = name
        if event_type is not MISSING:
            payload["event_type"] = try_enum_to_int(event_type)
        if trigger_metadata is not MISSING:
            payload["trigger_metadata"] = trigger_metadata.to_dict()
        if actions is not MISSING:
            if not actions:
                raise ValueError("At least one action must be provided.")
            for action in actions:
                if not isinstance(action, AutoModAction):
                    raise TypeError(
                        f"actions must be of type `AutoModAction` (or subtype), not {type(action)!r}"
                    )
            payload["actions"] = [a.to_dict() for a in actions]
        if enabled is not MISSING:
            payload["enabled"] = enabled
        if exempt_roles is not MISSING:
            payload["exempt_roles"] = (
                [e.id for e in exempt_roles] if exempt_roles is not None else []
            )
        if exempt_channels is not MISSING:
            payload["exempt_channels"] = (
                [e.id for e in exempt_channels] if exempt_channels is not None else []
            )

        data = await self.guild._state.http.edit_auto_moderation_rule(
            self.guild.id, self.id, reason=reason, **payload
        )
        return AutoModRule(data=data, guild=self.guild)

    async def delete(self, *, reason: Optional[str] = None) -> None:
        """|coro|

        Deletes the auto moderation rule.

        You must have :attr:`.Permissions.manage_guild` permission to do this.

        Parameters
        ----------
        reason: Optional[:class:`str`]
            The reason for deleting this rule. Shows up on the audit log.

        Raises
        ------
        Forbidden
            You do not have proper permissions to delete the rule.
        NotFound
            The rule does not exist.
        HTTPException
            Deleting the rule failed.
        """
        await self.guild._state.http.delete_auto_moderation_rule(
            self.guild.id, self.id, reason=reason
        )


class AutoModActionExecution:
    """Represents the data for an :func:`on_automod_action_execution` event.

    .. versionadded:: 2.6

    Attributes
    ----------
    action: Union[:class:`AutoModBlockMessageAction`, :class:`AutoModSendAlertAction`, :class:`AutoModTimeoutAction`, :class:`AutoModAction`]
        The action that was executed.
    guild: :class:`Guild`
        The guild this action was executed in.
    rule_id: :class:`int`
        The ID of the rule that matched.
    rule_trigger_type: :class:`AutoModTriggerType`
        The trigger type of the rule that matched.
    user_id: :class:`int`
        The ID of the user that triggered this action.
        See also :attr:`.user`.
    channel_id: Optional[:class:`int`]
        The channel or thread ID in which the event occurred, if any.
        See also :attr:`.channel`.
    message_id: Optional[:class:`int`]
        The ID of the message that matched. ``None`` if the message was blocked,
        or if the content was not part of a message.
        See also :attr:`.message`.
    alert_message_id: Optional[:class:`int`]
        The ID of the alert message sent as a result of this action, if any.
        See also :attr:`.alert_message`.
    content: :class:`str`
        The content that matched.

        Requires :attr:`Intents.message_content` to be enabled,
        otherwise this field will be empty.

    matched_keyword: Optional[:class:`str`]
        The keyword or regex that matched.

    matched_content: Optional[:class:`str`]
        The substring of :attr:`.content` that matched the rule/keyword.

        Requires :attr:`Intents.message_content` to be enabled,
        otherwise this field will be empty.
    """

    __slots__ = (
        "action",
        "guild",
        "rule_id",
        "rule_trigger_type",
        "user_id",
        "channel_id",
        "message_id",
        "alert_message_id",
        "content",
        "matched_keyword",
        "matched_content",
    )

    def __init__(self, *, data: AutoModerationActionExecutionEvent, guild: Guild) -> None:
        self.guild: Guild = guild
        self.action: AutoModAction = _automod_action_factory(data["action"])
        self.rule_id: int = int(data["rule_id"])
        self.rule_trigger_type: AutoModTriggerType = try_enum(
            AutoModTriggerType, data["rule_trigger_type"]
        )
        self.user_id: int = int(data["user_id"])
        self.channel_id: Optional[int] = _get_as_snowflake(data, "channel_id")
        self.message_id: Optional[int] = _get_as_snowflake(data, "message_id")
        self.alert_message_id: Optional[int] = _get_as_snowflake(data, "alert_system_message_id")
        self.content: str = data.get("content") or ""
        self.matched_keyword: Optional[str] = data.get("matched_keyword")
        self.matched_content: Optional[str] = data.get("matched_content")

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} guild={self.guild!r} action={self.action!r}"
            f" rule_id={self.rule_id!r} rule_trigger_type={self.rule_trigger_type!r}"
            f" channel={self.channel!r} user_id={self.user_id!r} message_id={self.message_id!r}"
            f" alert_message_id={self.alert_message_id!r} content={self.content!r}"
            f" matched_keyword={self.matched_keyword!r} matched_content={self.matched_content!r}>"
        )

    @property
    def user(self) -> Optional[Member]:
        """Optional[:class:`Member`]: The guild member that triggered this action.
        May be ``None`` if the member cannot be found. See also :attr:`.user_id`.
        """
        return self.guild.get_member(self.user_id)

    @property
    def channel(self) -> Optional[Union[GuildChannel, Thread]]:
        """Optional[Union[:class:`abc.GuildChannel`, :class:`Thread`]]:
        The channel or thread in which the event occurred, if any.
        """
        return self.guild._resolve_channel(self.channel_id)

    @property
    def message(self) -> Optional[Message]:
        """Optional[:class:`Message`]: The message that matched, if any.
        Not available if the message was blocked, if the content was not part of a message,
        or if the message was not found in the message cache.
        """
        return self.guild._state._get_message(self.message_id)

    @property
    def alert_message(self) -> Optional[Message]:
        """Optional[:class:`Message`]: The alert message sent as a result of this action, if any.
        Only available if :attr:`action.type <AutoModAction.type>` is :attr:`~AutoModActionType.send_alert_message`
        and the message was found in the message cache.
        """
        return self.guild._state._get_message(self.alert_message_id)


_action_map: Dict[int, Type[AutoModAction]] = {
    AutoModActionType.block_message.value: AutoModBlockMessageAction,
    AutoModActionType.send_alert_message.value: AutoModSendAlertAction,
    AutoModActionType.timeout.value: AutoModTimeoutAction,
}


def _automod_action_factory(data: AutoModActionPayload) -> AutoModAction:
    tp = _action_map.get(data["type"], AutoModAction)
    return tp._from_dict(data)



# SPDX-License-Identifier: MIT

"""disnake.types
~~~~~~~~~~~~~~

Typings for the Discord API

:copyright: (c) 2015-2021 Rapptz, 2021-present Disnake Development
:license: MIT, see LICENSE for more details.

"""



# SPDX-License-Identifier: MIT

from __future__ import annotations

import re

import pytest

from disnake import flags


class TestFlags(flags.BaseFlags):
    """A test class for flag testing."""

    __test__ = False

    @flags.flag_value
    def one(self):
        return 1 << 0

    @flags.flag_value
    def two(self):
        return 1 << 1

    @flags.flag_value
    def four(self):
        return 1 << 2

    @flags.alias_flag_value
    def three(self):
        return 1 << 0 | 1 << 1

    @flags.flag_value
    def sixteen(self):
        return 1 << 4


class OtherTestFlags(flags.BaseFlags):
    """Another test class for flag testing."""

    @flags.flag_value
    def other_one(self):
        return 1 << 0


def test_all_flags_value() -> None:
    assert flags.all_flags_value(TestFlags.VALID_FLAGS) == 0b10111


def test_flag_creation() -> None:
    assert TestFlags.VALID_FLAGS == {
        "one": 1,
        "two": 2,
        "three": 3,
        "four": 4,
        "sixteen": 16,
    }

    assert TestFlags.DEFAULT_VALUE == 0


def test_flag_creation_inverted() -> None:
    class InvertedFlags(flags.BaseFlags, inverted=True):
        @flags.flag_value
        def one(self):
            return 1 << 0

        @flags.flag_value
        def two(self):
            return 1 << 1

        @flags.flag_value
        def four(self):
            return 1 << 2

        @flags.alias_flag_value
        def three(self):
            return 1 << 0 | 1 << 1

        @flags.flag_value
        def sixteen(self):
            return 1 << 4

    assert InvertedFlags.VALID_FLAGS == {
        "one": 1,
        "two": 2,
        "three": 3,
        "four": 4,
        "sixteen": 16,
    }

    assert InvertedFlags.DEFAULT_VALUE == 0b10111


def test_flag_creation_empty() -> None:
    with pytest.raises(RuntimeError, match="At least one flag must be defined"):

        class InvertedFlags(flags.BaseFlags):
            not_a_flag = 2


class TestFlagValue:
    def test_flag_value_creation(self) -> None:
        flag = flags.flag_value(lambda x: 1 << 2)
        assert 1 << 2 == flag.flag

    def test_flag_value_or(self) -> None:
        ins = TestFlags.four | TestFlags.one

        assert isinstance(ins, TestFlags)
        assert ins.value == 5
        assert (TestFlags.two | ins).value == 7

        assert not ins.value & TestFlags.sixteen.flag
        ins |= TestFlags.sixteen
        assert ins.value == 21

        ins = TestFlags()
        assert ins.value == 0
        assert (ins | TestFlags.one).value == 1

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for |:")):
            _ = TestFlags.four | 32  # type: ignore

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for |:")):
            _ = 32 | TestFlags.four  # type: ignore

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for |:")):
            _ = TestFlags.four | OtherTestFlags.other_one  # type: ignore

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for |:")):
            _ = TestFlags.four | OtherTestFlags(other_one=True)  # type: ignore

    def test_flag_value_invert(self) -> None:
        ins = ~TestFlags.four
        assert isinstance(ins, TestFlags)

        assert ins.value == 23 - 4

    def test_flag_value_xor(self) -> None:
        ins = TestFlags(one=True, two=True)

        ins ^= TestFlags.two
        assert ins.value == 1

        ins ^= TestFlags.three
        assert ins.value == 2

        assert (ins ^ TestFlags.four).value == 6


class TestBaseFlags:
    def test__init__default_value(self) -> None:
        ins = TestFlags()
        assert ins.DEFAULT_VALUE is ins.value

    def test__init__kwargs(self) -> None:
        ins = TestFlags(one=True, two=False)
        assert ins.one is True
        assert ins.two is False

    def test__init__invalid_kwargs(self) -> None:
        with pytest.raises(TypeError, match="'h' is not a valid flag name."):
            TestFlags(h=True)

    def test_set_require_bool(self) -> None:
        with pytest.raises(TypeError, match="Value to set for TestFlags must be a bool."):
            TestFlags(one="h")  # type: ignore

        ins = TestFlags()

        with pytest.raises(TypeError, match="Value to set for TestFlags must be a bool."):
            ins.two = "h"  # type: ignore

    def test__eq__(self) -> None:
        ins = TestFlags(one=True, two=True)
        other = TestFlags(one=True, two=True)

        assert ins is not other
        assert ins == other
        assert not ins != other

        ins.two = False
        assert not ins == other
        assert ins != other

    def test__eq__flag_value(self) -> None:
        ins = TestFlags(one=True)
        other = TestFlags(one=True, two=True)

        assert ins == TestFlags.one
        assert TestFlags.one == ins

        assert not ins != TestFlags.one
        assert ins != TestFlags.two

        assert other != TestFlags.one
        assert other != TestFlags.two

        assert other == TestFlags.three

    def test__and__(self) -> None:
        ins = TestFlags(one=True, two=True)
        other = TestFlags(one=True, two=True)

        third = ins & other
        assert third is not ins
        assert third.value == 0b011
        assert third == ins == other

        ins.one = False
        third = ins & other
        assert third is not ins
        assert third.value == 0b010

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for &:")):
            _ = ins & "44"  # type: ignore

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for &:")):
            _ = "44" & ins  # type: ignore

    def test__iand__(self) -> None:
        ins = TestFlags(one=True, two=True)
        other = TestFlags(one=True, two=True)

        third = ins
        ins &= other
        assert third is ins
        assert ins.value == 0b011

        other.two = False
        other.four = True
        ins &= other
        assert third is ins
        assert ins.value == 0b001

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for &=:")):
            ins &= 14  # type: ignore

    def test__or__(self) -> None:
        ins = TestFlags(one=True, two=False)
        other = TestFlags(one=False, two=True)

        third = ins | other
        assert third is not ins
        assert ins.value == 0b001
        assert other.value == 0b010
        assert third.value == 0b011

        ins.one = False
        third = ins | other
        assert third.value == 0b010
        assert third is not ins

        ins.value = other.value
        third = ins | other
        assert third is not ins
        assert third.value == 0b10

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for |:")):
            _ = ins | 28  # type: ignore

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for |:")):
            _ = 28 | ins  # type: ignore

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for |:")):
            _ = ins | OtherTestFlags.other_one  # type: ignore

    def test__ior__(self) -> None:
        ins = TestFlags(one=True, two=False)
        other = TestFlags(one=False, two=True)

        third = ins
        ins |= other
        assert ins is third
        assert ins.value == 0b011
        assert other.value == 0b010

        other.four = True
        ins |= other
        assert ins.value == 0b111

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for |=:")):
            ins |= True  # type: ignore

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for |=:")):
            ins |= OtherTestFlags.other_one  # type: ignore

    def test__xor__(self) -> None:
        ins = TestFlags(one=True, two=False)
        other = TestFlags(one=False, two=True)

        third = ins ^ other
        assert third.value == 0b011
        assert third is not ins

        other.one = True
        third = ins ^ other
        assert third.value == 0b010

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for ^:")):
            _ = ins ^ "h"  # type: ignore

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for ^:")):
            _ = "h" ^ ins  # type: ignore

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for ^:")):
            _ = ins ^ OtherTestFlags.other_one  # type: ignore

    def test__ixor__(self) -> None:
        ins = TestFlags(one=True, two=False)
        other = TestFlags(one=False, two=True)

        third = ins
        ins ^= other
        assert ins is third
        assert ins.value == 0b011

        ins.two = False
        other.one = True
        ins ^= other
        assert ins.value == 0b010

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for ^=:")):
            ins ^= "stability"  # type: ignore

        with pytest.raises(TypeError, match=re.escape("unsupported operand type(s) for ^=:")):
            ins ^= OtherTestFlags.other_one  # type: ignore

    def test__le__(self) -> None:
        ins = TestFlags(one=True, two=False)
        other = TestFlags(one=False, two=True)

        assert not ins <= other
        other.one = True
        assert ins <= other

        with pytest.raises(
            TypeError, match="'<=' not supported between instances of 'TestFlags' and 'int'"
        ):
            _ = ins <= 4  # type: ignore

        other.value = ins.value
        assert ins <= other

    def test__ge__(self) -> None:
        ins = TestFlags(one=True, two=False)
        other = TestFlags(one=False, two=True)

        assert not ins >= other
        ins.two = True
        assert ins >= other

        with pytest.raises(
            TypeError, match="'>=' not supported between instances of 'TestFlags' and 'int'"
        ):
            _ = ins >= 4  # type: ignore

        other.value = ins.value
        assert ins >= other

    def test__lt__(self) -> None:
        ins = TestFlags(one=True, two=False)
        other = TestFlags(one=False, two=True)

        assert not ins < other
        other.one = True
        assert ins < other

        with pytest.raises(
            TypeError, match="'<' not supported between instances of 'TestFlags' and 'int'"
        ):
            _ = ins < 4  # type: ignore

        other.value = ins.value
        assert not ins < other

    def test__gt__(self) -> None:
        ins = TestFlags(one=True, two=False)
        other = TestFlags(one=False, two=True)

        assert not ins > other
        ins.two = True
        assert ins > other

        with pytest.raises(
            TypeError, match="'>' not supported between instances of 'TestFlags' and 'int'"
        ):
            _ = ins > 4  # type: ignore

        other.value = ins.value
        assert not ins > other

    def test__invert__(self) -> None:
        ins = TestFlags(one=True)
        assert ins.value == 0b0001
        other = ~ins
        # assert that invert does not modify anything in-place
        assert ins.value == 0b0001
        # the other `0` here is because invert does not invert values that are not defined
        # assert other.value == goal
        assert other.value == 0b10110

    def test__hash__(self) -> None:
        ins = TestFlags(one=True)
        assert hash(ins) == hash(ins.value)

    def test_iter(self) -> None:
        ins = TestFlags(one=True, two=False)

        assert len(list(iter(ins))) == 4

        for flag, value in iter(ins):
            assert flag in ins.VALID_FLAGS
            assert getattr(ins, flag) == value

    def test_from_value(self) -> None:
        ins = TestFlags._from_value(0b101)
        assert ins.value == 0b101

    def test_set_and_get_flag(self) -> None:
        ins = TestFlags()
        assert ins.DEFAULT_VALUE == ins.value

        ins.two = True
        assert ins.two is True
        assert ins.value == TestFlags.two.flag == 1 << 1

    def test_alias_flag_value(self) -> None:
        ins = TestFlags(three=True)
        assert ins.value == 0b11
        ins.three = False
        assert ins.value == 0

    def test_alias_priority(self) -> None:
        ins = TestFlags(three=False, two=True, one=False)
        assert ins.value == 0b10

        ins = TestFlags(three=True, two=False, one=False)
        assert ins.value == 0b00

        ins = TestFlags(three=True, two=False)
        assert ins.value == 0b01

        ins = TestFlags(two=False, three=True)
        assert ins.value == 0b11

        ins = TestFlags(two=True, three=False)
        assert ins.value == 0b00


class _ListFlags(flags.ListBaseFlags):
    @flags.flag_value
    def flag1(self):
        return 1 << 0

    @flags.flag_value
    def flag2(self):
        return 1 << 1

    @flags.flag_value
    def flag3(self):
        return 1 << 2


class TestListBaseFlags:
    @pytest.mark.parametrize(
        ("kwargs", "expected_value", "expected_values"),
        [
            ({}, 0, []),
            ({"flag2": True}, 2, [1]),
            ({"flag3": True, "flag2": False, "flag1": True}, 5, [0, 2]),
        ],
    )
    def test_init(self, kwargs, expected_value, expected_values) -> None:
        flags = _ListFlags(**kwargs)
        assert flags.value == expected_value
        assert flags.values == expected_values

    def test_class_vars(self) -> None:
        assert _ListFlags.DEFAULT_VALUE == 0
        assert _ListFlags.VALID_FLAGS == {"flag1": 1, "flag2": 2, "flag3": 4}

    def test_from_values(self) -> None:
        flags = _ListFlags._from_values([0, 2, 3, 4, 5, 10])
        assert flags.value == 1085
        assert flags.values == [0, 2, 3, 4, 5, 10]

    @pytest.mark.parametrize("value", [-10, -1, 100])
    def test_from_values__invalid(self, value) -> None:
        with pytest.raises(ValueError, match=re.escape("Flag values must be within [0, 64)")):
            _ListFlags._from_values([0, 2, value, 3])

    def test_update(self) -> None:
        flags = _ListFlags(flag2=True)
        flags.flag2 = False
        flags.flag3 = True
        assert flags.value == 4
        assert flags.values == [2]

    def test_operators(self) -> None:
        a = _ListFlags(flag1=True, flag3=True)
        b = _ListFlags(flag1=False, flag3=True)
        assert a.values == [0, 2]
        assert b.values == [2]
        assert (~(a & b)).values == [0, 1]


class TestIntents:
    def test_all_only_valid(self) -> None:
        """Test that Intents.all() doesn't include flags that aren't defined."""
        intents = flags.Intents.all()

        assert not (1 << 18 | 1 << 17) & intents.value
        assert 1 << 20 & intents.value



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import List, Literal, Optional, TypedDict

from .snowflake import Snowflake
from .user import PartialUser

TeamMembershipState = Literal[1, 2]
TeamMemberRole = Literal["admin", "developer", "read_only"]


class TeamMember(TypedDict):
    membership_state: TeamMembershipState
    team_id: Snowflake
    user: PartialUser
    role: TeamMemberRole


class Team(TypedDict):
    id: Snowflake
    name: str
    owner_user_id: Snowflake
    members: List[TeamMember]
    icon: Optional[str]



# SPDX-License-Identifier: MIT

from __future__ import annotations

import io
import os
from typing import TYPE_CHECKING, Any, Literal, Optional, Tuple, Union

import yarl

from . import utils
from .errors import DiscordException
from .file import File

__all__ = ("Asset",)

if TYPE_CHECKING:
    from typing_extensions import Self

    from .state import ConnectionState
    from .webhook.async_ import BaseWebhook, _WebhookState

    ValidStaticFormatTypes = Literal["webp", "jpeg", "jpg", "png"]
    ValidAssetFormatTypes = Literal["webp", "jpeg", "jpg", "png", "gif"]
    AnyState = Union[ConnectionState, _WebhookState[BaseWebhook]]

AssetBytes = Union[utils._BytesLike, "AssetMixin"]

VALID_STATIC_FORMATS = frozenset({"jpeg", "jpg", "webp", "png"})
VALID_ASSET_FORMATS = VALID_STATIC_FORMATS | {"gif"}


MISSING = utils.MISSING


class AssetMixin:
    url: str
    _state: Optional[AnyState]

    __slots__: Tuple[str, ...] = ("_state",)

    async def read(self) -> bytes:
        """|coro|

        Retrieves the content of this asset as a :class:`bytes` object.

        Raises
        ------
        DiscordException
            There was no internal connection state.
        HTTPException
            Downloading the asset failed.
        NotFound
            The asset was deleted.

        Returns
        -------
        :class:`bytes`
            The content of the asset.
        """
        if self._state is None:
            raise DiscordException("Invalid state (no ConnectionState provided)")

        return await self._state.http.get_from_cdn(self.url)

    async def save(
        self, fp: Union[str, bytes, os.PathLike, io.BufferedIOBase], *, seek_begin: bool = True
    ) -> int:
        """|coro|

        Saves this asset into a file-like object.

        Parameters
        ----------
        fp: Union[:class:`io.BufferedIOBase`, :class:`os.PathLike`]
            The file-like object to save this asset to or the filename
            to use. If a filename is passed then a file is created with that
            filename and used instead.
        seek_begin: :class:`bool`
            Whether to seek to the beginning of the file after saving is
            successfully done.

        Raises
        ------
        DiscordException
            There was no internal connection state.
        HTTPException
            Downloading the asset failed.
        NotFound
            The asset was deleted.

        Returns
        -------
        :class:`int`
            The number of bytes written.
        """
        data = await self.read()
        if isinstance(fp, io.BufferedIOBase):
            written = fp.write(data)
            if seek_begin:
                fp.seek(0)
            return written
        else:
            with open(fp, "wb") as f:
                return f.write(data)

    async def to_file(
        self,
        *,
        spoiler: bool = False,
        filename: Optional[str] = None,
        description: Optional[str] = None,
    ) -> File:
        """|coro|

        Converts the asset into a :class:`File` suitable for sending via
        :meth:`abc.Messageable.send`.

        .. versionadded:: 2.5

        .. versionchanged:: 2.6
            Raises :exc:`TypeError` instead of ``InvalidArgument``.

        Parameters
        ----------
        spoiler: :class:`bool`
            Whether the file is a spoiler.
        filename: Optional[:class:`str`]
            The filename to display when uploading to Discord. If this is not given, it defaults to
            the name of the asset's URL.
        description: Optional[:class:`str`]
            The file's description.

        Raises
        ------
        DiscordException
            The asset does not have an associated state.
        HTTPException
            Downloading the asset failed.
        NotFound
            The asset was deleted.
        TypeError
            The asset is a unicode emoji or a sticker with lottie type.

        Returns
        -------
        :class:`File`
            The asset as a file suitable for sending.
        """
        data = await self.read()

        if not filename:
            filename = yarl.URL(self.url).name
            # if the filename doesn't have an extension (e.g. widget member avatars),
            # try to infer it from the data
            if not os.path.splitext(filename)[1]:
                ext = utils._get_extension_for_data(data)
                if ext:
                    filename += ext

        return File(io.BytesIO(data), filename=filename, spoiler=spoiler, description=description)


class Asset(AssetMixin):
    """Represents a CDN asset on Discord.

    .. collapse:: operations

        .. describe:: str(x)

            Returns the URL of the CDN asset.

        .. describe:: len(x)

            Returns the length of the CDN asset's URL.

        .. describe:: x == y

            Checks if the asset is equal to another asset.

        .. describe:: x != y

            Checks if the asset is not equal to another asset.

        .. describe:: hash(x)

            Returns the hash of the asset.
    """

    __slots__: Tuple[str, ...] = (
        "_url",
        "_animated",
        "_key",
    )

    BASE = "https://cdn.discordapp.com"

    # only used in special cases where Discord doesn't provide an asset on the CDN url
    BASE_MEDIA = "https://media.discordapp.net"

    def __init__(self, state: AnyState, *, url: str, key: str, animated: bool = False) -> None:
        self._state: AnyState = state
        self._url: str = url
        self._animated: bool = animated
        self._key: str = key

    @classmethod
    def _from_default_avatar(cls, state: AnyState, index: int) -> Self:
        return cls(
            state,
            url=f"{cls.BASE}/embed/avatars/{index}.png",
            key=str(index),
            animated=False,
        )

    @classmethod
    def _from_avatar(cls, state: AnyState, user_id: int, avatar: str) -> Self:
        animated = avatar.startswith("a_")
        format = "gif" if animated else "png"
        return cls(
            state,
            url=f"{cls.BASE}/avatars/{user_id}/{avatar}.{format}?size=1024",
            key=avatar,
            animated=animated,
        )

    @classmethod
    def _from_guild_avatar(
        cls, state: AnyState, guild_id: int, member_id: int, avatar: str
    ) -> Self:
        animated = avatar.startswith("a_")
        format = "gif" if animated else "png"
        return cls(
            state,
            url=f"{cls.BASE}/guilds/{guild_id}/users/{member_id}/avatars/{avatar}.{format}?size=1024",
            key=avatar,
            animated=animated,
        )

    @classmethod
    def _from_guild_banner(
        cls, state: AnyState, guild_id: int, member_id: int, banner: str
    ) -> Self:
        animated = banner.startswith("a_")
        format = "gif" if animated else "png"
        return cls(
            state,
            url=f"{cls.BASE}/guilds/{guild_id}/users/{member_id}/banners/{banner}.{format}?size=1024",
            key=banner,
            animated=animated,
        )

    @classmethod
    def _from_icon(cls, state: AnyState, object_id: int, icon_hash: str, path: str) -> Self:
        return cls(
            state,
            url=f"{cls.BASE}/{path}-icons/{object_id}/{icon_hash}.png?size=1024",
            key=icon_hash,
            animated=False,
        )

    @classmethod
    def _from_cover_image(cls, state: AnyState, object_id: int, cover_image_hash: str) -> Self:
        return cls(
            state,
            url=f"{cls.BASE}/app-assets/{object_id}/store/{cover_image_hash}.png?size=1024",
            key=cover_image_hash,
            animated=False,
        )

    @classmethod
    def _from_guild_image(cls, state: AnyState, guild_id: int, image: str, path: str) -> Self:
        return cls(
            state,
            url=f"{cls.BASE}/{path}/{guild_id}/{image}.png?size=1024",
            key=image,
            animated=False,
        )

    @classmethod
    def _from_guild_icon(cls, state: AnyState, guild_id: int, icon_hash: str) -> Self:
        animated = icon_hash.startswith("a_")
        format = "gif" if animated else "png"
        return cls(
            state,
            url=f"{cls.BASE}/icons/{guild_id}/{icon_hash}.{format}?size=1024",
            key=icon_hash,
            animated=animated,
        )

    @classmethod
    def _from_sticker_banner(cls, state: AnyState, banner: int) -> Self:
        return cls(
            state,
            url=f"{cls.BASE}/app-assets/710982414301790216/store/{banner}.png",
            key=str(banner),
            animated=False,
        )

    @classmethod
    def _from_banner(cls, state: AnyState, id: int, banner_hash: str) -> Self:
        animated = banner_hash.startswith("a_")
        format = "gif" if animated else "png"
        return cls(
            state,
            url=f"{cls.BASE}/banners/{id}/{banner_hash}.{format}?size=1024",
            key=banner_hash,
            animated=animated,
        )

    @classmethod
    def _from_role_icon(cls, state: AnyState, role_id: int, icon_hash: str) -> Self:
        animated = icon_hash.startswith("a_")
        format = "gif" if animated else "png"
        return cls(
            state,
            url=f"{cls.BASE}/role-icons/{role_id}/{icon_hash}.{format}?size=1024",
            key=icon_hash,
            animated=animated,
        )

    @classmethod
    def _from_guild_scheduled_event_image(
        cls, state: AnyState, event_id: int, image_hash: str
    ) -> Self:
        return cls(
            state,
            url=f"{cls.BASE}/guild-events/{event_id}/{image_hash}.png?size=2048",
            key=image_hash,
            animated=False,
        )

    @classmethod
    def _from_avatar_decoration(cls, state: AnyState, avatar_decoration_asset: str) -> Self:
        animated = avatar_decoration_asset.startswith("a_")
        return cls(
            state,
            url=f"{cls.BASE}/avatar-decoration-presets/{avatar_decoration_asset}.png?size=1024",
            key=avatar_decoration_asset,
            animated=animated,
        )

    def __str__(self) -> str:
        return self._url

    def __len__(self) -> int:
        return len(self._url)

    def __repr__(self) -> str:
        shorten = self._url.replace(self.BASE, "")
        return f"<Asset url={shorten!r}>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Asset) and self._url == other._url

    def __hash__(self) -> int:
        return hash(self._url)

    @property
    def url(self) -> str:
        """:class:`str`: Returns the underlying URL of the asset."""
        return self._url

    @property
    def key(self) -> str:
        """:class:`str`: Returns the identifying key of the asset."""
        return self._key

    def is_animated(self) -> bool:
        """Whether the asset is animated.

        :return type: :class:`bool`
        """
        return self._animated

    def replace(
        self,
        *,
        size: int = MISSING,
        format: ValidAssetFormatTypes = MISSING,
        static_format: ValidStaticFormatTypes = MISSING,
    ) -> Asset:
        """Returns a new asset with the passed components replaced.

        .. versionchanged:: 2.6
            Raises :exc:`ValueError` instead of ``InvalidArgument``.

        Parameters
        ----------
        size: :class:`int`
            The new size of the asset.
        format: :class:`str`
            The new format to change it to. Must be either
            'webp', 'jpeg', 'jpg', 'png', or 'gif' if it's animated.
        static_format: :class:`str`
            The new format to change it to if the asset isn't animated.
            Must be either 'webp', 'jpeg', 'jpg', or 'png'.

        Raises
        ------
        ValueError
            An invalid size or format was passed.

        Returns
        -------
        :class:`Asset`
            The newly updated asset.
        """
        url = yarl.URL(self._url)
        path, _ = os.path.splitext(url.path)

        if format is not MISSING:
            if self._animated:
                if format not in VALID_ASSET_FORMATS:
                    raise ValueError(f"format must be one of {VALID_ASSET_FORMATS}")
            else:
                if format not in VALID_STATIC_FORMATS:
                    raise ValueError(f"format must be one of {VALID_STATIC_FORMATS}")
            url = url.with_path(f"{path}.{format}")

        if static_format is not MISSING and not self._animated:
            if static_format not in VALID_STATIC_FORMATS:
                raise ValueError(f"static_format must be one of {VALID_STATIC_FORMATS}")
            url = url.with_path(f"{path}.{static_format}")

        if size is not MISSING:
            if not utils.valid_icon_size(size):
                raise ValueError("size must be a power of 2 between 16 and 4096")
            url = url.with_query(size=size)
        else:
            url = url.with_query(url.raw_query_string)

        url_str = str(url)
        return Asset(state=self._state, url=url_str, key=self._key, animated=self._animated)

    def with_size(self, size: int, /) -> Asset:
        """Returns a new asset with the specified size.

        .. versionchanged:: 2.6
            Raises :exc:`ValueError` instead of ``InvalidArgument``.

        Parameters
        ----------
        size: :class:`int`
            The new size of the asset.

        Raises
        ------
        ValueError
            The asset had an invalid size.

        Returns
        -------
        :class:`Asset`
            The newly updated asset.
        """
        if not utils.valid_icon_size(size):
            raise ValueError("size must be a power of 2 between 16 and 4096")

        url = str(yarl.URL(self._url).with_query(size=size))
        return Asset(state=self._state, url=url, key=self._key, animated=self._animated)

    def with_format(self, format: ValidAssetFormatTypes, /) -> Asset:
        """Returns a new asset with the specified format.

        .. versionchanged:: 2.6
            Raises :exc:`ValueError` instead of ``InvalidArgument``.

        Parameters
        ----------
        format: :class:`str`
            The new format of the asset.

        Raises
        ------
        ValueError
            The asset had an invalid format.

        Returns
        -------
        :class:`Asset`
            The newly updated asset.
        """
        if self._animated:
            if format not in VALID_ASSET_FORMATS:
                raise ValueError(f"format must be one of {VALID_ASSET_FORMATS}")
        else:
            if format not in VALID_STATIC_FORMATS:
                raise ValueError(f"format must be one of {VALID_STATIC_FORMATS}")

        url = yarl.URL(self._url)
        path, _ = os.path.splitext(url.path)
        url_str = str(url.with_path(f"{path}.{format}").with_query(url.raw_query_string))
        return Asset(state=self._state, url=url_str, key=self._key, animated=self._animated)

    def with_static_format(self, format: ValidStaticFormatTypes, /) -> Asset:
        """Returns a new asset with the specified static format.

        This only changes the format if the underlying asset is
        not animated. Otherwise, the asset is not changed.

        .. versionchanged:: 2.6
            Raises :exc:`ValueError` instead of ``InvalidArgument``.

        Parameters
        ----------
        format: :class:`str`
            The new static format of the asset.

        Raises
        ------
        ValueError
            The asset had an invalid format.

        Returns
        -------
        :class:`Asset`
            The newly updated asset.
        """
        if self._animated:
            return self
        return self.with_format(format)



# SPDX-License-Identifier: MIT

from typing import TYPE_CHECKING, Any, Callable, Coroutine, TypeVar, Union

if TYPE_CHECKING:
    from disnake import ApplicationCommandInteraction

    from .cog import Cog
    from .context import Context
    from .errors import CommandError

T = TypeVar("T")

Coro = Coroutine[Any, Any, T]
MaybeCoro = Union[T, Coro[T]]
CoroFunc = Callable[..., Coro[Any]]

Check = Union[
    Callable[["Cog", "Context[Any]"], MaybeCoro[bool]], Callable[["Context[Any]"], MaybeCoro[bool]]
]
AppCheck = Union[
    Callable[["Cog", "ApplicationCommandInteraction"], MaybeCoro[bool]],
    Callable[["ApplicationCommandInteraction"], MaybeCoro[bool]],
]
Hook = Union[Callable[["Cog", "Context[Any]"], Coro[Any]], Callable[["Context[Any]"], Coro[Any]]]
Error = Union[
    Callable[["Cog", "Context[Any]", "CommandError"], Coro[Any]],
    Callable[["Context[Any]", "CommandError"], Coro[Any]],
]


# This is merely a tag type to avoid circular import issues.
# Yes, this is a terrible solution but ultimately it is the only solution.
class _BaseCommand:
    __slots__ = ()



# SPDX-License-Identifier: MIT
from __future__ import annotations

from typing import TYPE_CHECKING, FrozenSet, List, Optional, Union

from .enums import OnboardingPromptType, try_enum
from .mixins import Hashable

if TYPE_CHECKING:
    from .emoji import Emoji, PartialEmoji
    from .guild import Guild, GuildChannel
    from .role import Role
    from .types.onboarding import (
        Onboarding as OnboardingPayload,
        OnboardingPrompt as OnboardingPromptPayload,
        OnboardingPromptOption as OnboardingPromptOptionPayload,
    )

__all__ = (
    "Onboarding",
    "OnboardingPrompt",
    "OnboardingPromptOption",
)


class Onboarding:
    """Represents a guild onboarding object.

    .. versionadded:: 2.9

    Attributes
    ----------
    guild: :class:`Guild`
        The guild this onboarding is part of.
    prompts: List[:class:`OnboardingPrompt`]
        The prompts shown during onboarding and in community customization.
    enabled: :class:`bool`
        Whether onboarding is enabled.
    default_channel_ids: FrozenSet[:class:`int`]
        The IDs of the channels that will be shown to new members by default.
    """

    __slots__ = (
        "guild",
        "prompts",
        "enabled",
        "default_channel_ids",
    )

    def __init__(self, *, guild: Guild, data: OnboardingPayload) -> None:
        self.guild: Guild = guild
        self._from_data(data)

    def _from_data(self, data: OnboardingPayload) -> None:
        self.prompts: List[OnboardingPrompt] = [
            OnboardingPrompt(data=prompt, guild=self.guild) for prompt in data["prompts"]
        ]
        self.enabled: bool = data["enabled"]
        self.default_channel_ids: FrozenSet[int] = (
            frozenset(map(int, exempt_channels))
            if (exempt_channels := data["default_channel_ids"])
            else frozenset()
        )

    def __repr__(self) -> str:
        return (
            f"<Onboarding guild={self.guild!r} prompts={self.prompts!r} enabled={self.enabled!r}>"
        )

    @property
    def default_channels(self) -> List[GuildChannel]:
        """List[:class:`abc.GuildChannel`]: The list of channels that will be shown to new members by default."""
        return list(filter(None, map(self.guild.get_channel, self.default_channel_ids)))


class OnboardingPrompt(Hashable):
    """Represents an onboarding prompt.

    .. versionadded:: 2.9

    Attributes
    ----------
    id: :class:`int`
        The onboarding prompt's ID.
    type: :class:`OnboardingPromptType`
        The onboarding prompt's type.
    options: List[:class:`OnboardingPromptOption`]
        The onboarding prompt's options.
    title: :class:`str`
        The onboarding prompt's title.
    single_select: :class:`bool`
        Whether users are limited to selecting one option for the prompt.
    required: :class:`bool`
        Whether the prompt is required before a user completes the onboarding flow.
    in_onboarding: :class:`bool`
        Whether the prompt is present in the onboarding flow.
        If ``False``, the prompt will only appear in community customization.
    """

    __slots__ = (
        "guild",
        "id",
        "title",
        "options",
        "single_select",
        "required",
        "in_onboarding",
        "type",
    )

    def __init__(self, *, guild: Guild, data: OnboardingPromptPayload) -> None:
        self.guild = guild

        self.id: int = int(data["id"])
        self.title: str = data["title"]
        self.options: List[OnboardingPromptOption] = [
            OnboardingPromptOption(data=option, guild=guild) for option in data["options"]
        ]
        self.single_select: bool = data["single_select"]
        self.required: bool = data["required"]
        self.in_onboarding: bool = data["in_onboarding"]
        self.type: OnboardingPromptType = try_enum(OnboardingPromptType, data["type"])

    def __str__(self) -> str:
        return self.title

    def __repr__(self) -> str:
        return (
            f"<OnboardingPrompt id={self.id!r} title={self.title!r} options={self.options!r}"
            f" single_select={self.single_select!r} required={self.required!r}"
            f" in_onboarding={self.in_onboarding!r} type={self.type!r}>"
        )


class OnboardingPromptOption(Hashable):
    """Represents an onboarding prompt option.

    .. versionadded:: 2.9

    Attributes
    ----------
    id: :class:`int`
        The prompt option's ID.
    emoji: Optional[Union[:class:`PartialEmoji`, :class:`Emoji`, :class:`str`]]
        The prompt option's emoji.
    title: :class:`str`
        The prompt option's title.
    description: Optional[:class:`str`]
        The prompt option's description.
    role_ids: FrozenSet[:class:`int`]
        The IDs of the roles that will be added to the user when they select this option.
    channel_ids: FrozenSet[:class:`int`]
        The IDs of the channels that the user will see when they select this option.
    """

    __slots__ = ("id", "title", "description", "emoji", "guild", "role_ids", "channel_ids")

    def __init__(self, *, guild: Guild, data: OnboardingPromptOptionPayload) -> None:
        self.guild: Guild = guild

        self.id: int = int(data["id"])
        self.title: str = data["title"]
        self.description: Optional[str] = data["description"]
        self.role_ids: FrozenSet[int] = (
            frozenset(map(int, roles_ids)) if (roles_ids := data.get("role_ids")) else frozenset()
        )
        self.channel_ids: FrozenSet[int] = (
            frozenset(map(int, channels_ids))
            if (channels_ids := data.get("channel_ids"))
            else frozenset()
        )

        self.emoji: Optional[Union[Emoji, PartialEmoji, str]]
        if emoji_data := data.get("emoji"):
            self.emoji = guild._state.get_reaction_emoji(emoji_data)
        else:
            self.emoji = None

    def __str__(self) -> str:
        return self.title

    def __repr__(self) -> str:
        return (
            f"<OnboardingPromptOption id={self.id!r} title={self.title!r} "
            f"description={self.description!r} emoji={self.emoji!r}>"
        )

    @property
    def roles(self) -> List[Role]:
        """List[:class:`Role`]: A list of roles that will be added to the user when they select this option."""
        return list(filter(None, map(self.guild.get_role, self.role_ids)))

    @property
    def channels(self) -> List[GuildChannel]:
        """List[:class:`abc.GuildChannel`]: A list of channels that the user will see when they select this option."""
        return list(filter(None, map(self.guild.get_channel, self.channel_ids)))



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, SupportsInt, Union

from . import utils
from .mixins import Hashable

if TYPE_CHECKING:
    import datetime

    SupportsIntCast = Union[SupportsInt, str, bytes, bytearray]

__all__ = ("Object",)


class Object(Hashable):
    """Represents a generic Discord object.

    The purpose of this class is to allow you to create 'miniature'
    versions of data classes if you want to pass in just an ID. Most functions
    that take in a specific data class with an ID can also take in this class
    as a substitute instead. Note that even though this is the case, not all
    objects (if any) actually inherit from this class.

    There are also some cases where some websocket events are received
    in :issue-dpy:`strange order <21>` and when such events happened you would
    receive this class rather than the actual data class. These cases are
    extremely rare.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two objects are equal.

        .. describe:: x != y

            Checks if two objects are not equal.

        .. describe:: hash(x)

            Returns the object's hash.

    Attributes
    ----------
    id: :class:`int`
        The ID of the object.
    """

    def __init__(self, id: SupportsIntCast) -> None:
        try:
            id = int(id)
        except ValueError:
            raise TypeError(
                f"id parameter must be convertable to int not {id.__class__!r}"
            ) from None
        else:
            self.id = id

    def __repr__(self) -> str:
        return f"<Object id={self.id!r}>"

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns the snowflake's creation time in UTC."""
        return utils.snowflake_time(self.id)



# SPDX-License-Identifier: MIT

from __future__ import annotations

import struct
from typing import IO, TYPE_CHECKING, ClassVar, Generator, Optional, Tuple

from .errors import DiscordException

__all__ = (
    "OggError",
    "OggPage",
    "OggStream",
)


class OggError(DiscordException):
    """An exception that is thrown for Ogg stream parsing errors."""

    pass


# https://tools.ietf.org/html/rfc3533
# https://tools.ietf.org/html/rfc7845


class OggPage:
    _header: ClassVar[struct.Struct] = struct.Struct("<xBQIIIB")
    if TYPE_CHECKING:
        flag: int
        gran_pos: int
        serial: int
        pagenum: int
        crc: int
        segnum: int

    def __init__(self, stream: IO[bytes]) -> None:
        try:
            header = stream.read(struct.calcsize(self._header.format))

            (
                self.flag,
                self.gran_pos,
                self.serial,
                self.pagenum,
                self.crc,
                self.segnum,
            ) = self._header.unpack(header)

            self.segtable: bytes = stream.read(self.segnum)
            bodylen = sum(struct.unpack("B" * self.segnum, self.segtable))
            self.data: bytes = stream.read(bodylen)
        except Exception:
            raise OggError("bad data stream") from None

    def iter_packets(self) -> Generator[Tuple[bytes, bool], None, None]:
        packetlen = offset = 0
        partial = True

        for seg in self.segtable:
            if seg == 255:
                packetlen += 255
                partial = True
            else:
                packetlen += seg
                yield self.data[offset : offset + packetlen], True
                offset += packetlen
                packetlen = 0
                partial = False

        if partial:
            yield self.data[offset:], False


class OggStream:
    def __init__(self, stream: IO[bytes]) -> None:
        self.stream: IO[bytes] = stream

    def _next_page(self) -> Optional[OggPage]:
        head = self.stream.read(4)
        if head == b"OggS":
            return OggPage(self.stream)
        elif not head:
            return None
        else:
            raise OggError("invalid header magic")

    def _iter_pages(self) -> Generator[OggPage, None, None]:
        page = self._next_page()
        while page:
            yield page
            page = self._next_page()

    def iter_packets(self) -> Generator[bytes, None, None]:
        partial = b""
        for page in self._iter_pages():
            for data, complete in page.iter_packets():
                partial += data
                if complete:
                    yield partial
                    partial = b""



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from ...abc import Snowflake
from ...components import SelectOption, StringSelectMenu
from ...enums import ComponentType, SelectDefaultValueType
from ...utils import MISSING
from .base import BaseSelect, P, V_co, _create_decorator

if TYPE_CHECKING:
    from typing_extensions import Self

    from ...emoji import Emoji
    from ...partial_emoji import PartialEmoji
    from ..item import DecoratedItem, ItemCallbackType


__all__ = (
    "StringSelect",
    "Select",
    "string_select",
    "select",
)


SelectOptionInput = Union[List[SelectOption], List[str], Dict[str, str]]


def _parse_select_options(options: SelectOptionInput) -> List[SelectOption]:
    if isinstance(options, dict):
        return [SelectOption(label=key, value=val) for key, val in options.items()]

    return [opt if isinstance(opt, SelectOption) else SelectOption(label=opt) for opt in options]


class StringSelect(BaseSelect[StringSelectMenu, str, V_co]):
    """Represents a UI string select menu.

    This is usually represented as a drop down menu.

    In order to get the selected items that the user has chosen, use :attr:`.values`.

    .. versionadded:: 2.0

    .. versionchanged:: 2.7
        Renamed from ``Select`` to ``StringSelect``.

    Parameters
    ----------
    custom_id: :class:`str`
        The ID of the select menu that gets received during an interaction.
        If not given then one is generated for you.
    placeholder: Optional[:class:`str`]
        The placeholder text that is shown if nothing is selected, if any.
    min_values: :class:`int`
        The minimum number of items that must be chosen for this select menu.
        Defaults to 1 and must be between 1 and 25.
    max_values: :class:`int`
        The maximum number of items that must be chosen for this select menu.
        Defaults to 1 and must be between 1 and 25.
    options: Union[List[:class:`disnake.SelectOption`], List[:class:`str`], Dict[:class:`str`, :class:`str`]]
        A list of options that can be selected in this menu. Use explicit :class:`.SelectOption`\\s
        for fine-grained control over the options. Alternatively, a list of strings will be treated
        as a list of labels, and a dict will be treated as a mapping of labels to values.

        .. versionchanged:: 2.5
            Now also accepts a list of str or a dict of str to str, which are then appropriately parsed as
            :class:`.SelectOption` labels and values.

    disabled: :class:`bool`
        Whether the select is disabled.
    row: Optional[:class:`int`]
        The relative row this select menu belongs to. A Discord component can only have 5
        rows. By default, items are arranged automatically into those 5 rows. If you'd
        like to control the relative positioning of the row then passing an index is advised.
        For example, row=1 will show up before row=2. Defaults to ``None``, which is automatic
        ordering. The row number must be between 0 and 4 (i.e. zero indexed).

    Attributes
    ----------
    values: List[:class:`str`]
        A list of values that have been selected by the user.
    """

    __repr_attributes__: Tuple[str, ...] = BaseSelect.__repr_attributes__ + ("options",)

    # In practice this should never be used by anything, might as well have it anyway though.
    _default_value_type_map: ClassVar[
        Mapping[SelectDefaultValueType, Tuple[Type[Snowflake], ...]]
    ] = {}

    @overload
    def __init__(
        self: StringSelect[None],
        *,
        custom_id: str = ...,
        placeholder: Optional[str] = None,
        min_values: int = 1,
        max_values: int = 1,
        options: SelectOptionInput = ...,
        disabled: bool = False,
        row: Optional[int] = None,
    ) -> None:
        ...

    @overload
    def __init__(
        self: StringSelect[V_co],
        *,
        custom_id: str = ...,
        placeholder: Optional[str] = None,
        min_values: int = 1,
        max_values: int = 1,
        options: SelectOptionInput = ...,
        disabled: bool = False,
        row: Optional[int] = None,
    ) -> None:
        ...

    def __init__(
        self,
        *,
        custom_id: str = MISSING,
        placeholder: Optional[str] = None,
        min_values: int = 1,
        max_values: int = 1,
        options: SelectOptionInput = MISSING,
        disabled: bool = False,
        row: Optional[int] = None,
    ) -> None:
        super().__init__(
            StringSelectMenu,
            ComponentType.string_select,
            custom_id=custom_id,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            disabled=disabled,
            default_values=None,
            row=row,
        )
        self._underlying.options = [] if options is MISSING else _parse_select_options(options)

    @classmethod
    def from_component(cls, component: StringSelectMenu) -> Self:
        return cls(
            custom_id=component.custom_id,
            placeholder=component.placeholder,
            min_values=component.min_values,
            max_values=component.max_values,
            options=component.options,
            disabled=component.disabled,
            row=None,
        )

    @property
    def options(self) -> List[SelectOption]:
        """List[:class:`disnake.SelectOption`]: A list of options that can be selected in this select menu."""
        return self._underlying.options

    @options.setter
    def options(self, value: List[SelectOption]) -> None:
        if not isinstance(value, list):
            raise TypeError("options must be a list of SelectOption")
        if not all(isinstance(obj, SelectOption) for obj in value):
            raise TypeError("all list items must subclass SelectOption")

        self._underlying.options = value

    def add_option(
        self,
        *,
        label: str,
        value: str = MISSING,
        description: Optional[str] = None,
        emoji: Optional[Union[str, Emoji, PartialEmoji]] = None,
        default: bool = False,
    ) -> None:
        """Adds an option to the select menu.

        To append a pre-existing :class:`.SelectOption` use the
        :meth:`append_option` method instead.

        Parameters
        ----------
        label: :class:`str`
            The label of the option. This is displayed to users.
            Can only be up to 100 characters.
        value: :class:`str`
            The value of the option. This is not displayed to users.
            If not given, defaults to the label. Can only be up to 100 characters.
        description: Optional[:class:`str`]
            An additional description of the option, if any.
            Can only be up to 100 characters.
        emoji: Optional[Union[:class:`str`, :class:`.Emoji`, :class:`.PartialEmoji`]]
            The emoji of the option, if available. This can either be a string representing
            the custom or unicode emoji or an instance of :class:`.PartialEmoji` or :class:`.Emoji`.
        default: :class:`bool`
            Whether this option is selected by default.

        Raises
        ------
        ValueError
            The number of options exceeds 25.
        """
        option = SelectOption(
            label=label,
            value=value,
            description=description,
            emoji=emoji,
            default=default,
        )

        self.append_option(option)

    def append_option(self, option: SelectOption) -> None:
        """Appends an option to the select menu.

        Parameters
        ----------
        option: :class:`disnake.SelectOption`
            The option to append to the select menu.

        Raises
        ------
        ValueError
            The number of options exceeds 25.
        """
        if len(self._underlying.options) >= 25:
            raise ValueError("maximum number of options already provided")

        self._underlying.options.append(option)


Select = StringSelect  # backwards compatibility


S_co = TypeVar("S_co", bound="StringSelect", covariant=True)


@overload
def string_select(
    *,
    placeholder: Optional[str] = None,
    custom_id: str = ...,
    min_values: int = 1,
    max_values: int = 1,
    options: SelectOptionInput = ...,
    disabled: bool = False,
    row: Optional[int] = None,
) -> Callable[[ItemCallbackType[V_co, StringSelect[V_co]]], DecoratedItem[StringSelect[V_co]]]:
    ...


@overload
def string_select(
    cls: Callable[P, S_co], *_: P.args, **kwargs: P.kwargs
) -> Callable[[ItemCallbackType[V_co, S_co]], DecoratedItem[S_co]]:
    ...


def string_select(
    cls: Callable[..., S_co] = StringSelect[Any], **kwargs: Any
) -> Callable[[ItemCallbackType[V_co, S_co]], DecoratedItem[S_co]]:
    """A decorator that attaches a string select menu to a component.

    The function being decorated should have three parameters, ``self`` representing
    the :class:`disnake.ui.View`, the :class:`disnake.ui.StringSelect` that was
    interacted with, and the :class:`disnake.MessageInteraction`.

    In order to get the selected items that the user has chosen within the callback
    use :attr:`StringSelect.values`.

    .. versionchanged:: 2.7
        Renamed from ``select`` to ``string_select``.

    Parameters
    ----------
    cls: Callable[..., :class:`StringSelect`]
        A callable (may be a :class:`StringSelect` subclass) to create a new instance of this component.
        If provided, the other parameters described below do not apply.
        Instead, this decorator will accept the same keywords as the passed callable/class does.

        .. versionadded:: 2.6
    placeholder: Optional[:class:`str`]
        The placeholder text that is shown if nothing is selected, if any.
    custom_id: :class:`str`
        The ID of the select menu that gets received during an interaction.
        It is recommended not to set this parameter to prevent conflicts.
    row: Optional[:class:`int`]
        The relative row this select menu belongs to. A Discord component can only have 5
        rows. By default, items are arranged automatically into those 5 rows. If you'd
        like to control the relative positioning of the row then passing an index is advised.
        For example, row=1 will show up before row=2. Defaults to ``None``, which is automatic
        ordering. The row number must be between 0 and 4 (i.e. zero indexed).
    min_values: :class:`int`
        The minimum number of items that must be chosen for this select menu.
        Defaults to 1 and must be between 1 and 25.
    max_values: :class:`int`
        The maximum number of items that must be chosen for this select menu.
        Defaults to 1 and must be between 1 and 25.
    options: Union[List[:class:`disnake.SelectOption`], List[:class:`str`], Dict[:class:`str`, :class:`str`]]
        A list of options that can be selected in this menu. Use explicit :class:`.SelectOption`\\s
        for fine-grained control over the options. Alternatively, a list of strings will be treated
        as a list of labels, and a dict will be treated as a mapping of labels to values.

        .. versionchanged:: 2.5
            Now also accepts a list of str or a dict of str to str, which are then appropriately parsed as
            :class:`.SelectOption` labels and values.

    disabled: :class:`bool`
        Whether the select is disabled. Defaults to ``False``.
    """
    return _create_decorator(cls, **kwargs)


select = string_select  # backwards compatibility



# SPDX-License-Identifier: MIT

from __future__ import annotations

import unicodedata
from typing import TYPE_CHECKING, List, Literal, Optional, Tuple, Type, Union

from .asset import Asset, AssetMixin
from .enums import StickerFormatType, StickerType, try_enum
from .errors import InvalidData
from .mixins import Hashable
from .utils import MISSING, _get_as_snowflake, cached_slot_property, find, get, snowflake_time

__all__ = (
    "StickerPack",
    "StickerItem",
    "Sticker",
    "StandardSticker",
    "GuildSticker",
)

if TYPE_CHECKING:
    import datetime

    from .guild import Guild
    from .state import ConnectionState
    from .types.sticker import (
        EditGuildSticker,
        GuildSticker as GuildStickerPayload,
        ListStickerPacks as ListStickerPacksPayload,
        StandardSticker as StandardStickerPayload,
        Sticker as StickerPayload,
        StickerItem as StickerItemPayload,
        StickerPack as StickerPackPayload,
    )
    from .user import User


class StickerPack(Hashable):
    """Represents a sticker pack.

    .. versionadded:: 2.0

    .. versionchanged:: 2.8
        :attr:`cover_sticker_id`, :attr:`cover_sticker` and :attr:`banner` are now optional.

    .. collapse:: operations

        .. describe:: str(x)

            Returns the name of the sticker pack.

        .. describe:: x == y

           Checks if the sticker pack is equal to another sticker pack.

        .. describe:: x != y

           Checks if the sticker pack is not equal to another sticker pack.

    Attributes
    ----------
    name: :class:`str`
        The name of the sticker pack.
    description: :class:`str`
        The description of the sticker pack.
    id: :class:`int`
        The id of the sticker pack.
    stickers: List[:class:`StandardSticker`]
        The stickers of this sticker pack.
    sku_id: :class:`int`
        The SKU ID of the sticker pack.
    cover_sticker_id: Optional[:class:`int`]
         The ID of the sticker used for the cover of the sticker pack, if any.
    cover_sticker: Optional[:class:`StandardSticker`]
        The sticker used for the cover of the sticker pack, if any.
    """

    __slots__ = (
        "_state",
        "id",
        "stickers",
        "name",
        "sku_id",
        "cover_sticker_id",
        "cover_sticker",
        "description",
        "_banner",
    )

    def __init__(self, *, state: ConnectionState, data: StickerPackPayload) -> None:
        self._state: ConnectionState = state
        self._from_data(data)

    def _from_data(self, data: StickerPackPayload) -> None:
        self.id: int = int(data["id"])
        stickers = data["stickers"]
        self.stickers: List[StandardSticker] = [
            StandardSticker(state=self._state, data=sticker) for sticker in stickers
        ]
        self.name: str = data["name"]
        self.sku_id: int = int(data["sku_id"])
        self.cover_sticker_id: Optional[int] = _get_as_snowflake(data, "cover_sticker_id")
        self.cover_sticker: Optional[StandardSticker] = get(self.stickers, id=self.cover_sticker_id)
        self.description: str = data["description"]
        self._banner: Optional[int] = _get_as_snowflake(data, "banner_asset_id")

    @property
    def banner(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: The banner asset of the sticker pack, if any."""
        if not self._banner:
            return None
        return Asset._from_sticker_banner(self._state, self._banner)

    def __repr__(self) -> str:
        return f"<StickerPack id={self.id} name={self.name!r} description={self.description!r}>"

    def __str__(self) -> str:
        return self.name


class _StickerTag(Hashable, AssetMixin):
    __slots__ = ()

    id: int
    format: StickerFormatType

    @property
    def url(self) -> str:
        """:class:`str`: The url for the sticker's image."""
        # https://github.com/discord/discord-api-docs/issues/6675#issuecomment-1954755672
        base = Asset.BASE_MEDIA if self.format is StickerFormatType.gif else Asset.BASE
        return f"{base}/stickers/{self.id}.{self.format.file_extension}"

    async def read(self) -> bytes:
        """|coro|

        Retrieves the content of this sticker as a :class:`bytes` object.

        .. note::

            Stickers that use the :attr:`StickerFormatType.lottie` format cannot be read.

        Raises
        ------
        HTTPException
            Downloading the asset failed.
        NotFound
            The asset was deleted.
        TypeError
            The sticker is a lottie type.

        Returns
        -------
        :class:`bytes`
            The content of the asset.
        """
        if self.format is StickerFormatType.lottie:
            raise TypeError('Cannot read stickers of format "lottie".')
        return await super().read()


class StickerItem(_StickerTag):
    """Represents a sticker item.

    .. versionadded:: 2.0

    .. collapse:: operations

        .. describe:: str(x)

            Returns the name of the sticker item.

        .. describe:: x == y

           Checks if the sticker item is equal to another sticker item.

        .. describe:: x != y

           Checks if the sticker item is not equal to another sticker item.

    Attributes
    ----------
    name: :class:`str`
        The sticker's name.
    id: :class:`int`
        The ID of the sticker.
    format: :class:`StickerFormatType`
        The format for the sticker's image.
    """

    __slots__ = ("name", "id", "format")

    def __init__(self, *, state: ConnectionState, data: StickerItemPayload) -> None:
        self._state: ConnectionState = state
        self.name: str = data["name"]
        self.id: int = int(data["id"])
        self.format: StickerFormatType = try_enum(StickerFormatType, data["format_type"])

    def __repr__(self) -> str:
        return f"<StickerItem id={self.id} name={self.name!r} format={self.format}>"

    def __str__(self) -> str:
        return self.name

    async def fetch(self) -> Union[Sticker, StandardSticker, GuildSticker]:
        """|coro|

        Attempts to retrieve the full sticker data of the sticker item.

        Raises
        ------
        HTTPException
            Retrieving the sticker failed.

        Returns
        -------
        Union[:class:`StandardSticker`, :class:`GuildSticker`]
            The retrieved sticker.
        """
        data: StickerPayload = await self._state.http.get_sticker(self.id)
        cls, _ = _sticker_factory(data["type"])  # type: ignore
        return cls(state=self._state, data=data)


class Sticker(_StickerTag):
    """Represents a sticker.

    .. versionadded:: 1.6

    .. collapse:: operations

        .. describe:: str(x)

            Returns the name of the sticker.

        .. describe:: x == y

           Checks if the sticker is equal to another sticker.

        .. describe:: x != y

           Checks if the sticker is not equal to another sticker.

    Attributes
    ----------
    name: :class:`str`
        The sticker's name.
    id: :class:`int`
        The ID of the sticker.
    description: :class:`str`
        The description of the sticker.
    pack_id: :class:`int`
        The ID of the sticker's pack.
    format: :class:`StickerFormatType`
        The format for the sticker's image.
    """

    __slots__ = ("id", "name", "description", "format")

    def __init__(self, *, state: ConnectionState, data: StickerPayload) -> None:
        self._state: ConnectionState = state
        self._from_data(data)

    def _from_data(self, data: StickerPayload) -> None:
        self.id: int = int(data["id"])
        self.name: str = data["name"]
        self.description: str = data.get("description") or ""
        self.format: StickerFormatType = try_enum(StickerFormatType, data["format_type"])

    def __repr__(self) -> str:
        return f"<Sticker id={self.id} name={self.name!r}>"

    def __str__(self) -> str:
        return self.name

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns the sticker's creation time in UTC."""
        return snowflake_time(self.id)


class StandardSticker(Sticker):
    """Represents a sticker that is found in a standard sticker pack.

    .. versionadded:: 2.0

    .. collapse:: operations

        .. describe:: str(x)

            Returns the name of the sticker.

        .. describe:: x == y

           Checks if the sticker is equal to another sticker.

        .. describe:: x != y

           Checks if the sticker is not equal to another sticker.

    Attributes
    ----------
    name: :class:`str`
        The sticker's name.
    id: :class:`int`
        The ID of the sticker.
    description: :class:`str`
        The description of the sticker.
    pack_id: :class:`int`
        The ID of the sticker's pack.
    format: :class:`StickerFormatType`
        The format for the sticker's image.
    tags: List[:class:`str`]
        A list of tags for the sticker.
    sort_value: :class:`int`
        The sticker's sort order within its pack.
    """

    __slots__ = ("sort_value", "pack_id", "type", "tags")

    def _from_data(self, data: StandardStickerPayload) -> None:
        super()._from_data(data)
        self.sort_value: int = data["sort_value"]
        self.pack_id: int = int(data["pack_id"])
        self.type: StickerType = StickerType.standard

        try:
            self.tags: List[str] = [tag.strip() for tag in data["tags"].split(",")]
        except KeyError:
            self.tags = []

    def __repr__(self) -> str:
        return f"<StandardSticker id={self.id} name={self.name!r} pack_id={self.pack_id}>"

    async def pack(self) -> StickerPack:
        """|coro|

        Retrieves the sticker pack that this sticker belongs to.

        Raises
        ------
        InvalidData
            The corresponding sticker pack was not found.
        HTTPException
            Retrieving the sticker pack failed.

        Returns
        -------
        :class:`StickerPack`
            The retrieved sticker pack.
        """
        data: ListStickerPacksPayload = await self._state.http.list_sticker_packs()
        packs = data["sticker_packs"]
        pack = find(lambda d: int(d["id"]) == self.pack_id, packs)

        if pack:
            return StickerPack(state=self._state, data=pack)
        raise InvalidData(f"Could not find corresponding sticker pack for {self!r}")


class GuildSticker(Sticker):
    """Represents a sticker that belongs to a guild.

    .. versionadded:: 2.0

    .. collapse:: operations

        .. describe:: str(x)

            Returns the name of the sticker.

        .. describe:: x == y

           Checks if the sticker is equal to another sticker.

        .. describe:: x != y

           Checks if the sticker is not equal to another sticker.

    Attributes
    ----------
    name: :class:`str`
        The sticker's name.
    id: :class:`int`
        The ID of the sticker.
    description: :class:`str`
        The description of the sticker.
    format: :class:`StickerFormatType`
        The format for the sticker's image.
    available: :class:`bool`
        Whether this sticker is available for use.
    guild_id: :class:`int`
        The ID of the guild that this sticker is from.
    user: Optional[:class:`User`]
        The user that created this sticker. This can only be retrieved using
        :meth:`Guild.fetch_sticker`/:meth:`Guild.fetch_stickers` while
        having the :attr:`~Permissions.manage_guild_expressions` permission.
    emoji: :class:`str`
        The name of a unicode emoji that represents this sticker.
    """

    __slots__ = ("available", "guild_id", "user", "emoji", "type", "_cs_guild")

    def _from_data(self, data: GuildStickerPayload) -> None:
        super()._from_data(data)
        self.available: bool = data.get("available", True)
        self.guild_id: int = int(data["guild_id"])
        user = data.get("user")
        self.user: Optional[User] = self._state.store_user(user) if user else None
        self.emoji: str = data["tags"]
        self.type: StickerType = StickerType.guild

    def __repr__(self) -> str:
        return f"<GuildSticker name={self.name!r} id={self.id} guild_id={self.guild_id} user={self.user!r}>"

    @cached_slot_property("_cs_guild")
    def guild(self) -> Optional[Guild]:
        """Optional[:class:`Guild`]: The guild that this sticker is from.
        Could be ``None`` if the bot is not in the guild.

        .. versionadded:: 2.0
        """
        return self._state._get_guild(self.guild_id)

    async def edit(
        self,
        *,
        name: str = MISSING,
        description: Optional[str] = MISSING,
        emoji: str = MISSING,
        reason: Optional[str] = None,
    ) -> GuildSticker:
        """|coro|

        Edits a :class:`GuildSticker` for the guild.

        You must have :attr:`~Permissions.manage_guild_expressions` permission to
        do this.

        Parameters
        ----------
        name: :class:`str`
            The sticker's new name. Must be at least 2 characters.
        description: Optional[:class:`str`]
            The sticker's new description. Can be ``None``.
        emoji: :class:`str`
            The name of a unicode emoji that represents the sticker's expression.
        reason: Optional[:class:`str`]
            The reason for editing this sticker. Shows up on the audit log.

        Raises
        ------
        Forbidden
            You are not allowed to edit this sticker.
        HTTPException
            An error occurred editing the sticker.

        Returns
        -------
        :class:`GuildSticker`
            The newly modified sticker.
        """
        payload: EditGuildSticker = {}

        if name is not MISSING:
            payload["name"] = name

        if description is not MISSING:
            payload["description"] = description

        if emoji is not MISSING:
            try:
                emoji = unicodedata.name(emoji)
            except TypeError:
                pass
            else:
                emoji = emoji.replace(" ", "_")

            payload["tags"] = emoji

        data: GuildStickerPayload = await self._state.http.modify_guild_sticker(
            self.guild_id, self.id, payload, reason=reason
        )
        return GuildSticker(state=self._state, data=data)

    async def delete(self, *, reason: Optional[str] = None) -> None:
        """|coro|

        Deletes the custom :class:`Sticker` from the guild.

        You must have :attr:`~Permissions.manage_guild_expressions` permission to
        do this.

        Parameters
        ----------
        reason: Optional[:class:`str`]
            The reason for deleting this sticker. Shows up on the audit log.

        Raises
        ------
        Forbidden
            You are not allowed to delete this sticker.
        HTTPException
            An error occurred deleting the sticker.
        """
        await self._state.http.delete_guild_sticker(self.guild_id, self.id, reason=reason)


def _sticker_factory(
    sticker_type: Literal[1, 2]
) -> Tuple[Type[Union[StandardSticker, GuildSticker, Sticker]], StickerType]:
    value = try_enum(StickerType, sticker_type)
    if value == StickerType.standard:
        return StandardSticker, value
    elif value == StickerType.guild:
        return GuildSticker, value
    else:
        return Sticker, value



# SPDX-License-Identifier: MIT

from __future__ import annotations

import asyncio
import os
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Callable,
    ClassVar,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from ...components import AnySelectMenu, SelectDefaultValue
from ...enums import ComponentType, SelectDefaultValueType
from ...object import Object
from ...utils import MISSING, humanize_list
from ..item import DecoratedItem, Item

__all__ = ("BaseSelect",)

if TYPE_CHECKING:
    from typing_extensions import ParamSpec, Self

    from ...abc import Snowflake
    from ...interactions import MessageInteraction
    from ..item import ItemCallbackType
    from ..view import View

else:
    ParamSpec = TypeVar


S_co = TypeVar("S_co", bound="BaseSelect", covariant=True)
V_co = TypeVar("V_co", bound="Optional[View]", covariant=True)
SelectMenuT = TypeVar("SelectMenuT", bound=AnySelectMenu)
SelectValueT = TypeVar("SelectValueT")
P = ParamSpec("P")

SelectDefaultValueMultiInputType = Union[SelectValueT, SelectDefaultValue]
# almost the same as above, but with `Object`; used for selects where the type isn't ambiguous (i.e. all except mentionable select)
SelectDefaultValueInputType = Union[SelectDefaultValueMultiInputType[SelectValueT], Object]


class BaseSelect(Generic[SelectMenuT, SelectValueT, V_co], Item[V_co], ABC):
    """Represents an abstract UI select menu.

    This is usually represented as a drop down menu.

    This isn't meant to be used directly, instead use one of the concrete select menu types:

    - :class:`disnake.ui.StringSelect`
    - :class:`disnake.ui.UserSelect`
    - :class:`disnake.ui.RoleSelect`
    - :class:`disnake.ui.MentionableSelect`
    - :class:`disnake.ui.ChannelSelect`

    .. versionadded:: 2.7
    """

    __repr_attributes__: Tuple[str, ...] = (
        "placeholder",
        "min_values",
        "max_values",
        "disabled",
    )
    # We have to set this to MISSING in order to overwrite the abstract property from WrappedComponent
    _underlying: SelectMenuT = MISSING

    # Subclasses are expected to set this
    _default_value_type_map: ClassVar[Mapping[SelectDefaultValueType, Tuple[Type[Snowflake], ...]]]

    def __init__(
        self,
        underlying_type: Type[SelectMenuT],
        component_type: ComponentType,
        *,
        custom_id: str,
        placeholder: Optional[str],
        min_values: int,
        max_values: int,
        disabled: bool,
        default_values: Optional[Sequence[SelectDefaultValueInputType[SelectValueT]]],
        row: Optional[int],
    ) -> None:
        super().__init__()
        self._selected_values: List[SelectValueT] = []
        self._provided_custom_id = custom_id is not MISSING
        custom_id = os.urandom(16).hex() if custom_id is MISSING else custom_id
        self._underlying = underlying_type._raw_construct(
            custom_id=custom_id,
            type=component_type,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            disabled=disabled,
            default_values=self._transform_default_values(default_values) if default_values else [],
        )
        self.row = row

    @property
    def custom_id(self) -> str:
        """:class:`str`: The ID of the select menu that gets received during an interaction."""
        return self._underlying.custom_id

    @custom_id.setter
    def custom_id(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError("custom_id must be None or str")

        self._underlying.custom_id = value

    @property
    def placeholder(self) -> Optional[str]:
        """Optional[:class:`str`]: The placeholder text that is shown if nothing is selected, if any."""
        return self._underlying.placeholder

    @placeholder.setter
    def placeholder(self, value: Optional[str]) -> None:
        if value is not None and not isinstance(value, str):
            raise TypeError("placeholder must be None or str")

        self._underlying.placeholder = value

    @property
    def min_values(self) -> int:
        """:class:`int`: The minimum number of items that must be chosen for this select menu."""
        return self._underlying.min_values

    @min_values.setter
    def min_values(self, value: int) -> None:
        self._underlying.min_values = int(value)

    @property
    def max_values(self) -> int:
        """:class:`int`: The maximum number of items that must be chosen for this select menu."""
        return self._underlying.max_values

    @max_values.setter
    def max_values(self, value: int) -> None:
        self._underlying.max_values = int(value)

    @property
    def disabled(self) -> bool:
        """:class:`bool`: Whether the select menu is disabled."""
        return self._underlying.disabled

    @disabled.setter
    def disabled(self, value: bool) -> None:
        self._underlying.disabled = bool(value)

    @property
    def default_values(self) -> List[SelectDefaultValue]:
        """List[:class:`.SelectDefaultValue`]: The list of values that are selected by default.
        Only available for auto-populated select menus.
        """
        return self._underlying.default_values

    @default_values.setter
    def default_values(
        self, value: Optional[Sequence[SelectDefaultValueInputType[SelectValueT]]]
    ) -> None:
        self._underlying.default_values = self._transform_default_values(value) if value else []

    @property
    def values(self) -> List[SelectValueT]:
        return self._selected_values

    @property
    def width(self) -> int:
        return 5

    def refresh_component(self, component: SelectMenuT) -> None:
        self._underlying = component

    def refresh_state(self, interaction: MessageInteraction) -> None:
        self._selected_values = interaction.resolved_values  # type: ignore

    @classmethod
    @abstractmethod
    def from_component(cls, component: SelectMenuT) -> Self:
        raise NotImplementedError

    def is_dispatchable(self) -> bool:
        """Whether the select menu is dispatchable. This will always return ``True``.

        :return type: :class:`bool`
        """
        return True

    @classmethod
    def _transform_default_values(
        cls, values: Sequence[SelectDefaultValueInputType[SelectValueT]]
    ) -> List[SelectDefaultValue]:
        result: List[SelectDefaultValue] = []

        for value in values:
            # If we have a SelectDefaultValue, just use it as-is
            if isinstance(value, SelectDefaultValue):
                if value.type not in cls._default_value_type_map:
                    allowed_types = [str(t) for t in cls._default_value_type_map]
                    raise ValueError(
                        f"SelectDefaultValue.type should be {humanize_list(allowed_types, 'or')}, not {value.type}"
                    )
                result.append(value)
                continue

            # Otherwise, look through the list of allowed input types and
            # get the associated SelectDefaultValueType
            for (
                value_type,  # noqa: B007  # we use value_type outside of the loop
                types,
            ) in cls._default_value_type_map.items():
                if isinstance(value, types):
                    break
            else:
                allowed_types = [
                    t.__name__ for ts in cls._default_value_type_map.values() for t in ts
                ]
                allowed_types.append(SelectDefaultValue.__name__)
                raise TypeError(
                    f"Expected type of default value to be {humanize_list(allowed_types, 'or')}, not {type(value)!r}"
                )

            result.append(SelectDefaultValue(value.id, value_type))

        return result


def _create_decorator(
    # FIXME(3.0): rename `cls` parameter to more closely represent any callable argument type
    cls: Callable[P, S_co],
    /,
    *args: P.args,
    **kwargs: P.kwargs,
) -> Callable[[ItemCallbackType[V_co, S_co]], DecoratedItem[S_co]]:
    if args:
        # the `*args` def above is just to satisfy the typechecker
        raise RuntimeError("expected no *args")

    if not callable(cls):
        raise TypeError("cls argument must be callable")

    def decorator(func: ItemCallbackType[V_co, S_co]) -> DecoratedItem[S_co]:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("select function must be a coroutine function")

        func.__discord_ui_model_type__ = cls
        func.__discord_ui_model_kwargs__ = kwargs
        return func  # type: ignore

    return decorator
